"""
Chip manager for fixed target
This version changed to python3 March2020 by RLO
"""

import json
import re
import sys
from pathlib import Path
from pprint import pformat

import bluesky.plan_stubs as bps
import numpy as np
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.attenuator.attenuator import ReadOnlyAttenuator
from dodal.devices.i24.beamstop import Beamstop, BeamstopPositions
from dodal.devices.i24.dual_backlight import BacklightPositions, DualBacklight
from dodal.devices.i24.i24_detector_motion import DetectorMotion
from dodal.devices.i24.pmac import PMAC, EncReset, LaserSettings

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import (
    ChipType,
    Fiducials,
    MappingType,
)
from mx_bluesky.beamlines.i24.serial.log import (
    SSX_LOGGER,
    _read_visit_directory_from_file,
    log_on_entry,
)
from mx_bluesky.beamlines.i24.serial.parameters import (
    FixedTargetParameters,
    get_chip_format,
    get_chip_map,
)
from mx_bluesky.beamlines.i24.serial.parameters.constants import (
    CS_FILES_PATH,
    LITEMAP_PATH,
    PARAM_FILE_PATH_FT,
    PVAR_FILE_PATH,
)
from mx_bluesky.beamlines.i24.serial.setup_beamline import Pilatus, caget, caput, pv
from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector import (
    get_detector_type,
)

# An approximation of the chip size for the move during fiducials alignment.
CHIP_MOVES = {
    ChipType.Oxford: 25.40,
    ChipType.OxfordInner: 24.60,
    ChipType.Custom: 25.40,
    ChipType.Minichip: 25.40,
}
OXFORD_CHIP_WIDTH = 8
PVAR_TEMPLATE = f"P3%0{2}d1"
CHIPTYPE_PV = pv.me14e_gp1
MAPTYPE_PV = pv.me14e_gp2
NUM_EXPOSURES_PV = pv.me14e_gp3
PUMP_REPEAT_PV = pv.me14e_gp4
MAP_FILEPATH_PV = pv.me14e_gp5


@log_on_entry
def initialise_stages(
    pmac: PMAC = inject("pmac"),
) -> MsgGenerator:
    """Initialise the portable stages PVs, usually used only once right after setting \
        up the stages either after use at different facility.
    """
    group = "initialise_stages"
    SSX_LOGGER.info("Setting velocity, acceleration and limits for stages")

    yield from bps.abs_set(pmac.x.velocity, 20, group=group)
    yield from bps.abs_set(pmac.y.velocity, 20, group=group)
    yield from bps.abs_set(pmac.z.velocity, 20, group=group)
    yield from bps.abs_set(pmac.x.acceleration_time, 0.01, group=group)
    yield from bps.abs_set(pmac.y.acceleration_time, 0.01, group=group)
    yield from bps.abs_set(pmac.z.acceleration_time, 0.01, group=group)
    yield from bps.abs_set(pmac.x.high_limit_travel, 30, group=group)
    yield from bps.abs_set(pmac.x.low_limit_travel, -29, group=group)
    yield from bps.abs_set(pmac.y.high_limit_travel, 30, group=group)
    yield from bps.abs_set(pmac.y.low_limit_travel, -30, group=group)
    yield from bps.abs_set(pmac.z.high_limit_travel, 5.1, group=group)
    yield from bps.abs_set(pmac.z.low_limit_travel, -4.1, group=group)
    caput(CHIPTYPE_PV, 1)  # chip type
    caput(MAPTYPE_PV, 0)  # map type
    caput(NUM_EXPOSURES_PV, 1)  # num exposures
    caput(PUMP_REPEAT_PV, 0)  # pump repeat
    caput(pv.me14e_filepath, "test")
    caput(pv.me14e_chip_name, "albion")
    caput(pv.me14e_dcdetdist, 1480)
    caput(pv.me14e_exptime, 0.01)
    yield from bps.abs_set(pmac.enc_reset, EncReset.ENC5, group=group)
    yield from bps.abs_set(pmac.enc_reset, EncReset.ENC6, group=group)
    yield from bps.abs_set(pmac.enc_reset, EncReset.ENC7, group=group)
    yield from bps.abs_set(pmac.enc_reset, EncReset.ENC8, group=group)

    caput(pv.pilat_cbftemplate, 0)

    yield from bps.sleep(0.1)
    SSX_LOGGER.info("Clearing General Purpose PVs 1-120")
    for i in range(4, 120):
        if i == 100:
            # Do not clear visit PV
            continue
        pvar = "ME14E-MO-IOC-01:GP" + str(i)
        caput(pvar, 0)
        sys.stdout.write(".")
        sys.stdout.flush()

    SSX_LOGGER.info("Initialisation of the stages complete")
    yield from bps.wait(group=group)


@log_on_entry
def read_parameters(
    detector_stage: DetectorMotion,
    attenuator: ReadOnlyAttenuator,
) -> MsgGenerator:
    """ Read the parameters from user input and create the parameter model for a fixed \
        target collection.

    Args:
        detector_stage (DetectorMotion): The detector stage device.
        attenuator (ReadOnlyAttenuator): A read-only attenuator device to get the \
            transmission value.

    Returns:
        FixedTargetParameters: Parameter model for fixed target collections

    """
    SSX_LOGGER.info("Creating parameter model from input.")

    filename = caget(pv.me14e_chip_name)
    det_type = yield from get_detector_type(detector_stage)
    chip_params = get_chip_format(ChipType(int(caget(CHIPTYPE_PV))))
    map_type = int(caget(MAPTYPE_PV))
    if map_type == MappingType.Lite and chip_params.chip_type in [
        ChipType.Oxford,
        ChipType.OxfordInner,
    ]:
        chip_map = get_chip_map()
    else:
        chip_map = []
    pump_repeat = int(caget(PUMP_REPEAT_PV))

    # If file name ends in a digit this causes processing/pilatus pain.
    # Append an underscore
    if isinstance(det_type, Pilatus):
        caput(pv.pilat_cbftemplate, 0)
        m = re.search(r"\d+$", filename)
        if m is not None:
            # Note for future reference. Appending underscore causes more hassle and
            # high probability of users accidentally overwriting data. Use a dash
            filename = filename + "-"
            SSX_LOGGER.debug(
                f"Requested filename ends in a number. Appended dash: {filename}"
            )

    transmission = yield from bps.rd(attenuator.actual_transmission)

    params_dict = {
        "visit": _read_visit_directory_from_file().as_posix(),  # noqa
        "directory": caget(pv.me14e_filepath),
        "filename": filename,
        "exposure_time_s": caget(pv.me14e_exptime),
        "detector_distance_mm": caget(pv.me14e_dcdetdist),
        "detector_name": str(det_type),
        "num_exposures": int(caget(NUM_EXPOSURES_PV)),
        "transmission": transmission,
        "chip": chip_params.model_dump(),
        "map_type": map_type,
        "pump_repeat": pump_repeat,
        "checker_pattern": bool(caget(pv.me14e_gp111)),
        "chip_map": chip_map,
        "laser_dwell_s": float(caget(pv.me14e_gp103)) if pump_repeat != 0 else 0.0,
        "laser_delay_s": float(caget(pv.me14e_gp110)) if pump_repeat != 0 else 0.0,
        "pre_pump_exposure_s": float(caget(pv.me14e_gp109))
        if pump_repeat != 0
        else None,
    }

    SSX_LOGGER.info("Parameters for I24 serial collection: \n")
    SSX_LOGGER.info(pformat(params_dict))

    yield from bps.null()
    return FixedTargetParameters(**params_dict)


def scrape_pvar_file(fid: str, pvar_dir: Path = PVAR_FILE_PATH):
    block_start_list = []

    with open(pvar_dir / fid) as f:
        lines = f.readlines()
    for line in lines:
        line = line.rstrip()
        if line.startswith("#"):
            continue
        elif line.startswith("P3000"):
            continue
        elif line.startswith("P3011"):
            continue
        elif not len(line.split(" ")) == 2:
            continue
        else:
            entry = line.split(" ")
            block_num = entry[0][2:4]
            x = entry[0].split("=")[1]
            y = entry[1].split("=")[1]
            block_start_list.append([block_num, x, y])
    return block_start_list


@log_on_entry
def define_current_chip(
    chipid: str = "oxford",
    pmac: PMAC = inject("pmac"),
) -> MsgGenerator:
    SSX_LOGGER.debug("Run load stock map for just the first block")
    yield from load_stock_map("Just The First Block")
    """
    Not sure what this is for:
    print 'Setting Mapping Type to Lite'
    caput(pv.me14e_gp2, 1)
    """
    chip_type = int(caget(CHIPTYPE_PV))
    SSX_LOGGER.info(f"Chip type:{chip_type} Chipid:{chipid}")
    if chipid == "oxford":
        caput(CHIPTYPE_PV, 0)

    with open(PVAR_FILE_PATH / f"{chipid}.pvar") as f:
        SSX_LOGGER.info(f"Opening {chipid}.pvar")
        for line in f.readlines():
            if line.startswith("#"):
                continue
            line_from_file = line.rstrip("\n")
            SSX_LOGGER.info(f"{line_from_file}")
            yield from bps.abs_set(pmac.pmac_string, line_from_file)


@log_on_entry
def upload_chip_map_to_geobrick(pmac: PMAC, chip_map: list[int]) -> MsgGenerator:
    """Upload the map parameters for an Oxford-type chip (width=8) to the geobrick.

    Args:
        pmac (PMAC): The PMAC device.
        chip_map (list[int]): A list of selected blocks to be collected.

    """
    SSX_LOGGER.info("Uploading Parameters for Oxford Chip to the GeoBrick")
    SSX_LOGGER.info(f"Chipid {ChipType.Oxford}, width {OXFORD_CHIP_WIDTH}")
    for block in range(1, 65):
        value = 1 if block in chip_map else 0
        pvar = PVAR_TEMPLATE % block
        pvar_str = f"{pvar}={value}"
        SSX_LOGGER.debug(f"Set {pvar_str} for block {block}")
        yield from bps.abs_set(pmac.pmac_string, pvar_str, wait=True)
        # Wait for PMAC to be done processing PVAR string
        yield from bps.sleep(0.02)
    SSX_LOGGER.debug("Upload parameters done.")


@log_on_entry
def load_stock_map(map_choice: str = "clear") -> MsgGenerator:
    # TODO See https://github.com/DiamondLightSource/mx_bluesky/issues/122
    SSX_LOGGER.info("Adjusting Lite Map EDM Screen")
    SSX_LOGGER.debug("Please wait, adjusting lite map")
    #
    r33 = [19, 18, 17, 26, 31, 32, 33, 24, 25]
    r55 = [9, 10, 11, 12, 13, 16, 27, 30, 41, 40, 39, 38, 37, 34, 23, 20] + r33
    r77 = [
        7,
        6,
        5,
        4,
        3,
        2,
        1,
        14,
        15,
        28,
        29,
        42,
        43,
        44,
        45,
        46,
        47,
        48,
        49,
        36,
        35,
        22,
        21,
        8,
    ] + r55
    #
    h33 = [3, 2, 1, 6, 7, 8, 9, 4, 5]
    x33 = [31, 32, 33, 40, 51, 50, 49, 42, 41]
    x55 = [25, 24, 23, 22, 21, 34, 39, 52, 57, 58, 59, 60, 61, 48, 43, 30] + x33
    x77 = [
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        20,
        35,
        38,
        53,
        56,
        71,
        70,
        69,
        68,
        67,
        66,
        65,
        62,
        47,
        44,
        29,
        26,
    ] + x55
    x99 = [
        9,
        8,
        7,
        6,
        5,
        4,
        3,
        2,
        1,
        18,
        19,
        36,
        37,
        54,
        55,
        72,
        73,
        74,
        75,
        76,
        77,
        78,
        79,
        80,
        81,
        64,
        63,
        46,
        45,
        28,
        27,
        10,
    ] + x77  # noqa
    x44 = [22, 21, 20, 19, 30, 35, 46, 45, 44, 43, 38, 27, 28, 29, 36, 37]
    x49 = [x + 1 for x in range(49)]
    x66 = [
        10,
        11,
        12,
        13,
        14,
        15,
        18,
        31,
        34,
        47,
        50,
        51,
        52,
        53,
        54,
        55,
        42,
        39,
        26,
        23,
    ] + x44
    x88 = [
        8,
        7,
        6,
        5,
        4,
        3,
        2,
        1,
        16,
        17,
        32,
        33,
        48,
        49,
        64,
        63,
        62,
        61,
        60,
        59,
        58,
        57,
        56,
        41,
        40,
        25,
        24,
        9,
    ] + x66
    #
    # Columns for doing half chips
    c1 = [1, 2, 3, 4, 5, 6, 7, 8]
    c2 = [9, 10, 11, 12, 13, 14, 15, 16]
    c3 = [17, 18, 19, 20, 21, 22, 23, 24]
    c4 = [25, 26, 27, 28, 29, 30, 31, 32]
    c5 = [33, 34, 35, 36, 37, 38, 39, 40]
    c6 = [41, 42, 43, 44, 45, 46, 47, 48]
    c7 = [49, 50, 51, 52, 53, 54, 55, 56]
    c8 = [57, 58, 59, 60, 61, 62, 63, 64]
    half1 = c1 + c2 + c3 + c4
    half2 = c5 + c6 + c7 + c8

    map_dict = {}
    map_dict["Just The First Block"] = [1]
    map_dict["clear"] = []
    #
    map_dict["r33"] = r33
    map_dict["r55"] = r55
    map_dict["r77"] = r77
    #
    map_dict["h33"] = h33
    #
    map_dict["x33"] = x33
    map_dict["x44"] = x44
    map_dict["x49"] = x49
    map_dict["x55"] = x55
    map_dict["x66"] = x66
    map_dict["x77"] = x77
    map_dict["x88"] = x88
    map_dict["x99"] = x99

    map_dict["half1"] = half1
    map_dict["half2"] = half2

    SSX_LOGGER.info("Clearing GP 10-74")  # Actually 11-44
    for i in range(1, 65):
        pvar = "ME14E-MO-IOC-01:GP" + str(i + 10)
        caput(pvar, 0)
        sys.stdout.write(".")
        sys.stdout.flush()
    SSX_LOGGER.info("Map cleared")
    SSX_LOGGER.info(f"Loading Map Choice {map_choice}")
    for i in map_dict[map_choice]:
        pvar = "ME14E-MO-IOC-01:GP" + str(i + 10)
        caput(pvar, 1)
    SSX_LOGGER.debug("Load stock map done.")
    yield from bps.null()


@log_on_entry
def load_lite_map() -> MsgGenerator:
    SSX_LOGGER.debug("Run load stock map with 'clear' setting.")
    yield from load_stock_map("clear")
    # fmt: off
    # Oxford_block_dict is wrong (columns and rows need to flip) added in script below to generate it automatically however kept this for backwards compatiability/reference
    oxford_block_dict = {   # noqa: F841
        'A1': '01', 'A2': '02', 'A3': '03', 'A4': '04', 'A5': '05', 'A6': '06', 'A7': '07', 'A8': '08',
        'B1': '16', 'B2': '15', 'B3': '14', 'B4': '13', 'B5': '12', 'B6': '11', 'B7': '10', 'B8': '09',
        'C1': '17', 'C2': '18', 'C3': '19', 'C4': '20', 'C5': '21', 'C6': '22', 'C7': '23', 'C8': '24',
        'D1': '32', 'D2': '31', 'D3': '30', 'D4': '29', 'D5': '28', 'D6': '27', 'D7': '26', 'D8': '25',
        'E1': '33', 'E2': '34', 'E3': '35', 'E4': '36', 'E5': '37', 'E6': '38', 'E7': '39', 'E8': '40',
        'F1': '48', 'F2': '47', 'F3': '46', 'F4': '45', 'F5': '44', 'F6': '43', 'F7': '42', 'F8': '41',
        'G1': '49', 'G2': '50', 'G3': '51', 'G4': '52', 'G5': '53', 'G6': '54', 'G7': '55', 'G8': '56',
        'H1': '64', 'H2': '63', 'H3': '62', 'H4': '61', 'H5': '60', 'H6': '59', 'H7': '58', 'H8': '57',
    }
    # fmt: on
    chip_type = int(caget(CHIPTYPE_PV))
    if chip_type in [ChipType.Oxford, ChipType.OxfordInner]:
        SSX_LOGGER.info("Oxford Block Order")
        rows = ["A", "B", "C", "D", "E", "F", "G", "H"]
        columns = list(range(1, 10))
        btn_names = {}
        flip = True
        for x, column in enumerate(columns):
            for y, row in enumerate(rows):
                i = x * 8 + y
                if i % 8 == 0 and flip is False:
                    flip = True
                    z = 8 - (y + 1)
                elif i % 8 == 0 and flip is True:
                    flip = False
                    z = y
                elif flip is False:
                    z = y
                elif flip is True:
                    z = 8 - (y + 1)
                else:
                    SSX_LOGGER.warning("Problem in Chip Grid Creation")
                    break
                button_name = str(row) + str(column)
                lab_num = x * 8 + z
                label = f"{lab_num + 1:02d}"
                btn_names[button_name] = label
        block_dict = btn_names
    else:
        raise ValueError(f"{chip_type=} unrecognised")

    litemap_fid = f"{caget(MAP_FILEPATH_PV)}.lite"
    SSX_LOGGER.info("Please wait, loading LITE map")
    SSX_LOGGER.debug("Loading Lite Map")
    SSX_LOGGER.info("Opening %s" % (LITEMAP_PATH / litemap_fid))
    with open(LITEMAP_PATH / litemap_fid) as fh:
        f = fh.readlines()
    for line in f:
        entry = line.split()
        block_name = entry[0]
        yesno = entry[1]
        block_num = block_dict[block_name]
        pvar = "ME14E-MO-IOC-01:GP" + str(int(block_num) + 10)
        SSX_LOGGER.info(f"Block: {block_name} \tScanned: {yesno} \tPVAR: {pvar}")
    SSX_LOGGER.debug("Load lite map done")
    yield from bps.null()


@log_on_entry
def moveto(place: str = "origin", pmac: PMAC = inject("pmac")) -> MsgGenerator:
    SSX_LOGGER.info(f"Move to: {place}")
    if place == Fiducials.zero:
        SSX_LOGGER.info("Chip aspecific move.")
        yield from bps.trigger(pmac.to_xyz_zero)
        return

    chip_type = ChipType(int(caget(CHIPTYPE_PV)))
    SSX_LOGGER.info(f"Chip type is {chip_type}")
    if chip_type not in list(ChipType):
        SSX_LOGGER.warning("Unknown chip_type move")
        return

    SSX_LOGGER.info(f"{str(chip_type)} Move")
    chip_move = CHIP_MOVES[chip_type]

    if place == Fiducials.origin:
        yield from bps.mv(pmac.x, 0.0, pmac.y, 0.0)
    if place == Fiducials.fid1:
        yield from bps.mv(pmac.x, chip_move, pmac.y, 0.0)
    if place == Fiducials.fid2:
        yield from bps.mv(pmac.x, 0.0, pmac.y, chip_move)


@log_on_entry
def moveto_preset(
    place: str,
    pmac: PMAC = inject("pmac"),
    beamstop: Beamstop = inject("beamstop"),
    backlight: DualBacklight = inject("backlight"),
    det_stage: DetectorMotion = inject("detector_motion"),
) -> MsgGenerator:
    # Non Chip Specific Move
    if place == "zero":
        SSX_LOGGER.info(f"Moving to {place}")
        yield from bps.trigger(pmac.to_xyz_zero)

    elif place == "load_position":
        SSX_LOGGER.info("load position")
        yield from bps.abs_set(
            beamstop.pos_select, BeamstopPositions.ROBOT, group=place
        )
        yield from bps.abs_set(backlight, BacklightPositions.OUT, group=place)
        yield from bps.abs_set(det_stage.z, 1300, group=place)
        yield from bps.wait(group=place)

    elif place == "collect_position":
        SSX_LOGGER.info("collect position")
        caput(pv.me14e_filter, 20)
        yield from bps.mv(pmac.x, 0.0, pmac.y, 0.0, pmac.z, 0.0)
        yield from bps.abs_set(
            beamstop.pos_select, BeamstopPositions.DATA_COLLECTION, group=place
        )
        yield from bps.abs_set(backlight, BacklightPositions.IN, group=place)
        yield from bps.wait(group=place)

    elif place == "microdrop_position":
        SSX_LOGGER.info("microdrop align position")
        yield from bps.mv(pmac.x, 6.0, pmac.y, -7.8, pmac.z, 0.0)


@log_on_entry
def laser_control(laser_setting: str, pmac: PMAC = inject("pmac")) -> MsgGenerator:
    SSX_LOGGER.info(f"Move to: {laser_setting}")
    if laser_setting == "laser1on":  # these are in laser edm
        SSX_LOGGER.info("Laser 1 /BNC2 shutter is open")
        # Use M712 = 0 if triggering on falling edge. M712 =1 if on rising edge
        # Be sure to also change laser1off
        # caput(pv.me14e_pmac_str, ' M712=0 M711=1')
        yield from bps.abs_set(pmac.laser, LaserSettings.LASER_1_ON, wait=True)

    elif laser_setting == "laser1off":
        SSX_LOGGER.info("Laser 1 shutter is closed")
        yield from bps.abs_set(pmac.laser, LaserSettings.LASER_1_OFF, wait=True)

    elif laser_setting == "laser2on":
        SSX_LOGGER.info("Laser 2 / BNC3 shutter is open")
        yield from bps.abs_set(pmac.laser, LaserSettings.LASER_2_ON, wait=True)

    elif laser_setting == "laser2off":
        SSX_LOGGER.info("Laser 2 shutter is closed")
        yield from bps.abs_set(pmac.laser, LaserSettings.LASER_2_OFF, wait=True)

    elif laser_setting == "laser1burn":
        led_burn_time = caget(pv.me14e_gp103)
        SSX_LOGGER.info("Laser 1  on")
        SSX_LOGGER.info(f"Burn time is {led_burn_time} s")
        yield from bps.abs_set(pmac.laser, LaserSettings.LASER_1_ON, wait=True)
        yield from bps.sleep(float(led_burn_time))
        SSX_LOGGER.info("Laser 1 off")
        yield from bps.abs_set(pmac.laser, LaserSettings.LASER_1_OFF, wait=True)

    elif laser_setting == "laser2burn":
        led_burn_time = caget(pv.me14e_gp109)
        SSX_LOGGER.info("Laser 2 on")
        SSX_LOGGER.info(f"burntime {led_burn_time} s")
        yield from bps.abs_set(pmac.laser, LaserSettings.LASER_2_ON, wait=True)
        yield from bps.sleep(float(led_burn_time))
        SSX_LOGGER.info("Laser 2 off")
        yield from bps.abs_set(pmac.laser, LaserSettings.LASER_2_OFF, wait=True)


@log_on_entry
def scrape_mtr_directions(motor_file_path: Path = CS_FILES_PATH):
    with open(motor_file_path / "motor_direction.txt") as f:
        lines = f.readlines()
    mtr1_dir, mtr2_dir, mtr3_dir = 1.0, 1.0, 1.0
    for line in lines:
        if line.startswith("mtr1"):
            mtr1_dir = float(line.split("=")[1])
        elif line.startswith("mtr2"):
            mtr2_dir = float(line.split("=")[1])
        elif line.startswith("mtr3"):
            mtr3_dir = float(line.split("=")[1])
        else:
            continue
    SSX_LOGGER.debug(f"mt1_dir {mtr1_dir} mtr2_dir {mtr2_dir} mtr3_dir {mtr3_dir}")
    return mtr1_dir, mtr2_dir, mtr3_dir


@log_on_entry
def fiducial(point: int = 1, pmac: PMAC = inject("pmac")) -> MsgGenerator:
    scale = 10000.0  # noqa: F841

    mtr1_dir, mtr2_dir, mtr3_dir = scrape_mtr_directions(CS_FILES_PATH)

    rbv_1 = yield from bps.rd(pmac.x.user_readback)
    rbv_2 = yield from bps.rd(pmac.y.user_readback)
    rbv_3 = yield from bps.rd(pmac.z.user_readback)

    output_param_path = PARAM_FILE_PATH_FT
    output_param_path.mkdir(parents=True, exist_ok=True)
    SSX_LOGGER.info(f"Writing Fiducial File {output_param_path}/fiducial_{point}.txt")
    SSX_LOGGER.info("MTR\tRBV\tRAW\tCorr\tf_value")
    SSX_LOGGER.info(f"MTR1\t{rbv_1:1.4f}\t{mtr1_dir:f}")
    SSX_LOGGER.info(f"MTR2\t{rbv_2:1.4f}\t{mtr2_dir:f}")
    SSX_LOGGER.info(f"MTR3\t{rbv_3:1.4f}\t{mtr3_dir:f}")

    with open(output_param_path / f"fiducial_{point}.txt", "w") as f:
        f.write("MTR\tRBV\tCorr\n")
        f.write(f"MTR1\t{rbv_1:1.4f}\t{mtr1_dir:f}\n")
        f.write(f"MTR2\t{rbv_2:1.4f}\t{mtr2_dir:f}\n")
        f.write(f"MTR3\t{rbv_3:1.4f}\t{mtr3_dir:f}")
    SSX_LOGGER.info(f"Fiducial {point} set.")
    yield from bps.null()


def scrape_mtr_fiducials(
    point: int, param_path: Path = PARAM_FILE_PATH_FT
) -> tuple[float, float, float]:
    with open(param_path / f"fiducial_{point}.txt") as f:
        f_lines = f.readlines()[1:]
    f_x = float(f_lines[0].rsplit()[1])
    f_y = float(f_lines[1].rsplit()[1])
    f_z = float(f_lines[2].rsplit()[1])
    return f_x, f_y, f_z


@log_on_entry
def cs_maker(pmac: PMAC = inject("pmac")) -> MsgGenerator:
    """
    Coordinate system.

    Values for scalex, scaley, scalez, and skew, as well as the sign of
    Sx, Sy, Sz are stored in a .json file and should be modified there.
    Location of file: src/mx_bluesky/i24/serial/parameters/cs_maker.json

    Theory
    Rx: rotation about X-axis, pitch
    Ry: rotation about Y-axis, yaw
    Rz: rotation about Z-axis, roll
    The order of rotation is Roll->Yaw->Pitch (Rx*Ry*Rz)
    Rx           Ry          Rz
    |1  0   0| | Cy  0 Sy| |Cz -Sz 0|   | CyCz        -CxSz         Sy  |
    |0 Cx -Sx|*|  0  1  0|*|Sz  Cz 0| = | SxSyCz+CxSz -SxSySz+CxCz -SxCy|
    |0 Sx  Cx| |-Sy  0 Cy| | 0   0 1|   |-CxSyCz+SxSz  CxSySz+SxCz  CxCy|

    BELOW iS TEST TEST (CLOCKWISE)
    Rx           Ry          Rz
    |1  0   0| | Cy 0 -Sy| |Cz  Sz 0|   | CyCz         CxSz         -Sy |
    |0 Cx  Sx|*|  0  1  0|*|-Sz Cz 0| = | SxSyCz-CxSz  SxSySz+CxCz  SxCy|
    |0 -Sx Cx| | Sy  0 Cy| | 0   0 1|   | CxSyCz+SxSz  CxSySz-SxCz  CxCy|


    Skew:
    Skew is the difference between the Sz1 and Sz2 after rotation is taken out.
    This should be measured in situ prior to expriment, ie. measure by hand using
    opposite and adjacent RBV after calibration of scale factors.
    """
    chip_type = int(caget(CHIPTYPE_PV))
    fiducial_dict = {}
    fiducial_dict[0] = [25.400, 25.400]
    fiducial_dict[1] = [24.600, 24.600]
    fiducial_dict[2] = [25.400, 25.400]
    fiducial_dict[3] = [18.25, 18.25]
    SSX_LOGGER.info(f"Chip type is {chip_type} with size {fiducial_dict[chip_type]}")

    mtr1_dir, mtr2_dir, mtr3_dir = scrape_mtr_directions()
    f1_x, f1_y, f1_z = scrape_mtr_fiducials(1)
    f2_x, f2_y, f2_z = scrape_mtr_fiducials(2)
    SSX_LOGGER.info(f"mtr1 direction: {mtr1_dir}")
    SSX_LOGGER.info(f"mtr2 direction: {mtr2_dir}")
    SSX_LOGGER.info(f"mtr3 direction: {mtr3_dir}")

    # Scale parameters saved in json file
    try:
        with open(CS_FILES_PATH / "cs_maker.json") as fh:
            cs_info = json.load(fh)
    except json.JSONDecodeError:
        SSX_LOGGER.error("Invalid JSON file.")
        raise

    try:
        scalex, scaley, scalez = (
            float(cs_info["scalex"]),
            float(cs_info["scaley"]),
            float(cs_info["scalez"]),
        )
        skew = float(cs_info["skew"])
        Sx_dir, Sy_dir, Sz_dir = (
            int(cs_info["Sx_dir"]),
            int(cs_info["Sy_dir"]),
            int(cs_info["Sz_dir"]),
        )
    except KeyError:
        SSX_LOGGER.error("Wrong or missing key in the cs json file.")
        raise

    def check_dir(val):
        if val not in [1, -1]:
            raise ValueError("Wrong value for direction. Please set to either -1 or 1.")

    check_dir(Sx_dir)
    check_dir(Sy_dir)
    check_dir(Sz_dir)

    # Rotation Around Z
    # If stages upsidedown (I24) change sign of Sz
    Sz1 = -1 * f1_y / fiducial_dict[chip_type][0]
    Sz2 = f2_x / fiducial_dict[chip_type][1]
    Sz = Sz_dir * ((Sz1 + Sz2) / 2)
    Cz = np.sqrt(1 - Sz**2)
    SSX_LOGGER.info(f"Sz1 , {Sz1:1.4f}, {np.degrees(np.arcsin(Sz1)):1.4f}")
    SSX_LOGGER.info(f"Sz2 , {Sz2:1.4f}, {np.degrees(np.arcsin(Sz2)):1.4f}")
    SSX_LOGGER.info(f"Sz , {Sz:1.4f}, {np.degrees(np.arcsin(Sz)):1.4f}")
    SSX_LOGGER.info(f"Cz , {Cz:1.4f}, {np.degrees(np.arcsin(Cz)):1.4f}")
    # Rotation Around Y
    Sy = Sy_dir * f1_z / fiducial_dict[chip_type][0]
    Cy = np.sqrt(1 - Sy**2)
    SSX_LOGGER.info(f"Sy , {Sy:1.4f}, {np.degrees(np.arcsin(Sy)):1.4f}")
    SSX_LOGGER.info(f"Cy , {Cy:1.4f}, {np.degrees(np.arcsin(Cy)):1.4f}")
    # Rotation Around X
    # If stages upsidedown (I24) change sign of Sx
    Sx = Sx_dir * f2_z / fiducial_dict[chip_type][1]
    Cx = np.sqrt(1 - Sx**2)
    SSX_LOGGER.info(f"Sx , {Sx:1.4f}, {np.degrees(np.arcsin(Sx)):1.4f}")
    SSX_LOGGER.info(f"Cx , {Cx:1.4f}, {np.degrees(np.arcsin(Cx)):1.4f}")

    x1factor = mtr1_dir * scalex * (Cy * Cz)
    y1factor = mtr2_dir * scaley * (-1.0 * Cx * Sz)
    z1factor = mtr3_dir * scalez * Sy

    x2factor = mtr1_dir * scalex * ((Sx * Sy * Cz) + (Cx * Sz))
    y2factor = mtr2_dir * scaley * ((Cx * Cz) - (Sx * Sy * Sz))
    z2factor = mtr3_dir * scalez * (-1.0 * Sx * Cy)

    x3factor = mtr1_dir * scalex * ((Sx * Sz) - (Cx * Sy * Cz))
    y3factor = mtr2_dir * scaley * ((Cx * Sy * Sz) + (Sx * Cz))
    z3factor = mtr3_dir * scalez * (Cx * Cy)

    SSX_LOGGER.info(f"Skew being used is: {skew:1.4f}")
    s1 = np.degrees(np.arcsin(Sz1))
    s2 = np.degrees(np.arcsin(Sz2))
    rot = np.degrees(np.arcsin((Sz1 + Sz2) / 2))
    calc_skew = (s1 - rot) - (s2 - rot)
    SSX_LOGGER.info(f"s1:{s1:1.4f} s2:{s2:1.4f} rot:{rot:1.4f}")
    SSX_LOGGER.info(f"Calculated rotation from current fiducials is: {rot:1.4f}")
    SSX_LOGGER.info(f"Calculated Skew from current fiducials is: {calc_skew:1.4f}")
    SSX_LOGGER.info("Calculated Skew has been known to have the wrong sign")

    sinD = np.sin((skew / 2) * (np.pi / 180))
    cosD = np.cos((skew / 2) * (np.pi / 180))
    new_x1factor = (x1factor * cosD) + (y1factor * sinD)
    new_y1factor = (x1factor * sinD) + (y1factor * cosD)
    new_x2factor = (x2factor * cosD) + (y2factor * sinD)
    new_y2factor = (x2factor * sinD) + (y2factor * cosD)

    cs1 = f"#1->{new_x1factor:+1.3f}X{new_y1factor:+1.3f}Y{z1factor:+1.3f}Z"
    cs2 = f"#2->{new_x2factor:+1.3f}X{new_y2factor:+1.3f}Y{z2factor:+1.3f}Z"
    cs3 = f"#3->{x3factor:+1.3f}X{y3factor:+1.3f}Y{z3factor:+1.3f}Z"
    SSX_LOGGER.info(f"PMAC strings. \ncs1: {cs1} \ncs2: {cs2}cs3: {cs3}")
    SSX_LOGGER.info(
        """These next values should be 1.
        This is the sum of the squares of the factors divided by their scale."""
    )
    sqfact1 = np.sqrt(x1factor**2 + y1factor**2 + z1factor**2) / scalex
    sqfact2 = np.sqrt(x2factor**2 + y2factor**2 + z2factor**2) / scaley
    sqfact3 = np.sqrt(x3factor**2 + y3factor**2 + z3factor**2) / scalez
    SSX_LOGGER.info(f"{sqfact1:1.4f} \n {sqfact2:1.4f} \n {sqfact3:1.4f}")
    SSX_LOGGER.debug("Long wait, please be patient")
    yield from bps.trigger(pmac.to_xyz_zero)
    yield from bps.sleep(2.5)
    yield from set_pmac_strings_for_cs(pmac, {"cs1": cs1, "cs2": cs2, "cs3": cs3})
    yield from bps.trigger(pmac.to_xyz_zero)
    yield from bps.sleep(2.5)
    yield from bps.trigger(pmac.home, wait=True)
    yield from bps.sleep(2.5)
    SSX_LOGGER.debug(f"Chip_type is {chip_type}")
    if chip_type == 0:
        yield from bps.abs_set(pmac.pmac_string, "!x0.4y0.4", wait=True)
        yield from bps.sleep(2.5)
        yield from bps.trigger(pmac.home, wait=True)
    else:
        yield from bps.trigger(pmac.home, wait=True)
    SSX_LOGGER.debug("CSmaker done.")
    yield from bps.null()


def cs_reset(pmac: PMAC = inject("pmac")) -> MsgGenerator:
    """Used to clear CS when using Custom Chip"""
    cs1 = "#1->10000X+0Y+0Z"
    cs2 = "#2->+0X-10000Y+0Z"
    cs3 = "#3->0X+0Y-10000Z"
    strg = "\n".join([cs1, cs2, cs3])
    print(strg)
    yield from set_pmac_strings_for_cs(pmac, {"cs1": cs1, "cs2": cs2, "cs3": cs3})
    SSX_LOGGER.debug("CSreset Done")
    yield from bps.null()


def set_pmac_strings_for_cs(pmac: PMAC, cs_str: dict):
    """ A plan to set the pmac_string for the (x,y,z) axes while making or resetting \
        the coordinate system.

    Args:
        pmac (PMAC): PMAC device
        cs_str (dict): A dictionary containing a string for each axis, in the format: \
            {
                "cs1": "#1->1X+0Y+0Z",
                "cs2": "#2->...",
                "cs3": "#3->...",
            }

    Note. On the PMAC the axes allocations are: #1 - X, #2 - Y, #3 - Z.
    """
    yield from bps.abs_set(pmac.pmac_string, "&2", wait=True)
    yield from bps.abs_set(pmac.pmac_string, cs_str["cs1"], wait=True)
    yield from bps.abs_set(pmac.pmac_string, cs_str["cs2"], wait=True)
    yield from bps.abs_set(pmac.pmac_string, cs_str["cs3"], wait=True)


@log_on_entry
def pumpprobe_calc() -> MsgGenerator:
    # TODO See https://github.com/DiamondLightSource/mx_bluesky/issues/122
    SSX_LOGGER.info("Calculate and show exposure and dwell time for each option.")
    exptime = float(caget(pv.me14e_exptime))
    pumpexptime = float(caget(pv.me14e_gp103))
    movetime = 0.008
    SSX_LOGGER.info(f"X-ray exposure time {exptime}")
    SSX_LOGGER.info(f"Laser dwell time {pumpexptime}")
    repeat1 = 2 * 20 * (movetime + (pumpexptime + exptime) / 2)
    repeat2 = 4 * 20 * (movetime + (pumpexptime + exptime) / 2)
    repeat3 = 6 * 20 * (movetime + (pumpexptime + exptime) / 2)
    repeat5 = 10 * 20 * (movetime + (pumpexptime + exptime) / 2)
    repeat10 = 20 * 20 * (movetime + (pumpexptime + exptime) / 2)
    for pv_name, repeat in (
        (pv.me14e_gp104, repeat1),
        (pv.me14e_gp105, repeat2),
        (pv.me14e_gp106, repeat3),
        (pv.me14e_gp107, repeat5),
        (pv.me14e_gp108, repeat10),
    ):
        rounded = round(repeat, 4)
        caput(pv_name, rounded)
        SSX_LOGGER.info(f"Repeat ({pv_name}): {rounded} s")
    SSX_LOGGER.debug("PP calculations done")
    yield from bps.null()


@log_on_entry
def block_check(pmac: PMAC = inject("pmac")) -> MsgGenerator:
    # TODO See https://github.com/DiamondLightSource/mx_bluesky/issues/117
    caput(pv.me14e_gp9, 0)
    while True:
        if int(caget(pv.me14e_gp9)) == 0:
            chip_type = int(caget(CHIPTYPE_PV))
            if chip_type == ChipType.Minichip:
                SSX_LOGGER.info("Oxford mini chip in use.")
                block_start_list = scrape_pvar_file("minichip_oxford.pvar")
            elif chip_type == ChipType.Custom:
                SSX_LOGGER.error("This is a custom chip, no block check available!")
                raise ValueError(
                    "Chip type set to 'custom', which has no block check."
                    "If not using a custom chip, please double check chip in the GUI."
                )
            else:
                SSX_LOGGER.warning("Default is Oxford chip block start list.")
                block_start_list = scrape_pvar_file("oxford.pvar")
            for entry in block_start_list:
                if int(caget(pv.me14e_gp9)) != 0:
                    SSX_LOGGER.warning("Block Check Aborted")
                    yield from bps.sleep(1.0)
                    break
                block, x, y = entry
                SSX_LOGGER.debug(f"Block: {block} -> (x={x} y={y})")
                yield from bps.abs_set(pmac.pmac_string, f"!x{x}y{y}", wait=True)
                yield from bps.sleep(0.4)
        else:
            SSX_LOGGER.warning("Block Check Aborted due to GP 9 not equalling 0")
            break
        break
    SSX_LOGGER.debug("Block check done")
    yield from bps.null()



from pathlib import Path

from deepdiff.diff import DeepDiff
from pydantic_extra_types.semantic_version import SemanticVersion

from mx_bluesky.common.parameters.components import (
    PARAMETER_VERSION,
    TopNByMaxCountSelection,
)
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    AGAMEMNON_URL,
    AgamemnonLoadCentreCollect,
    SinglePin,
    _get_parameters_from_url,
    get_pin_type_from_agamemnon_parameters,
    populate_parameters_from_agamemnon,
)
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadThenCentre

EXPECTED_ROBOT_LOAD_AND_CENTRE_PARAMS = RobotLoadThenCentre(
    parameter_model_version=SemanticVersion.validate_from_str(str(PARAMETER_VERSION)),
    sample_id=12345,
    sample_puck=1,
    sample_pin=1,
    visit="cm00000-0",
    detector_distance_mm=180.8,
    storage_directory="/dls/tmp/data/year/cm00000-0/auto/test/xraycentring",
    snapshot_directory=Path("/dls/tmp/data/year/cm00000-0/auto/test/snapshots"),
    file_name="test_xtal",
    demand_energy_ev=12700.045934258673,
)

EXPECTED_PARAMETERS = AgamemnonLoadCentreCollect(
    visit="cm00000-0",
    detector_distance_mm=180.8,
    sample_id=12345,
    sample_puck=1,
    sample_pin=1,
    parameter_model_version=SemanticVersion.validate_from_str(str(PARAMETER_VERSION)),
    select_centres=TopNByMaxCountSelection(n=1),
    robot_load_then_centre=EXPECTED_ROBOT_LOAD_AND_CENTRE_PARAMS,
    demand_energy_ev=12700.045934258673,
)


def test_given_test_agamemnon_instruction_then_returns_none_loop_type():
    params = _get_parameters_from_url(AGAMEMNON_URL + "/example/collect")
    loop_type = get_pin_type_from_agamemnon_parameters(params)
    assert loop_type == SinglePin()


def test_given_test_agamemnon_instruction_then_load_centre_collect_parameters_populated():
    params = _get_parameters_from_url(AGAMEMNON_URL + "/example/collect")
    load_centre_collect = populate_parameters_from_agamemnon(params)
    difference = DeepDiff(
        load_centre_collect,
        EXPECTED_PARAMETERS,
    )
    assert not difference



from __future__ import annotations

from math import isclose
from typing import cast

import pydantic
from blueapi.core import BlueskyContext
from bluesky import plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.dcm import DCM
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import PandAFastGridScan, ZebraFastGridScan
from dodal.devices.flux import Flux
from dodal.devices.focusing_mirror import FocusingMirrorWithStripes, MirrorVoltages
from dodal.devices.i03.beamstop import Beamstop
from dodal.devices.motors import XYZPositioner
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot, SampleLocation
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.thawer import Thawer
from dodal.devices.undulator import Undulator
from dodal.devices.undulator_dcm import UndulatorDCM
from dodal.devices.webcam import Webcam
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from dodal.log import LOGGER
from ophyd_async.fastcs.panda import HDFPanda

from mx_bluesky.common.parameters.constants import OavConstants
from mx_bluesky.hyperion.device_setup_plans.utils import (
    fill_in_energy_if_not_supplied,
    start_preparing_data_collection_then_do_plan,
)
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan import (
    pin_centre_then_flyscan_plan,
)
from mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy import (
    RobotLoadAndEnergyChangeComposite,
    pin_already_loaded,
    robot_load_and_change_energy_plan,
)
from mx_bluesky.hyperion.experiment_plans.set_energy_plan import (
    SetEnergyComposite,
    set_energy_plan,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadThenCentre


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class RobotLoadThenCentreComposite:
    # common fields
    xbpm_feedback: XBPMFeedback
    attenuator: BinaryFilterAttenuator

    # GridDetectThenXRayCentreComposite fields
    aperture_scatterguard: ApertureScatterguard
    backlight: Backlight
    detector_motion: DetectorMotion
    eiger: EigerDetector
    zebra_fast_grid_scan: ZebraFastGridScan
    flux: Flux
    oav: OAV
    pin_tip_detection: PinTipDetection
    smargon: Smargon
    synchrotron: Synchrotron
    s4_slit_gaps: S4SlitGaps
    undulator: Undulator
    zebra: Zebra
    zocalo: ZocaloResults
    panda: HDFPanda
    panda_fast_grid_scan: PandAFastGridScan
    thawer: Thawer
    sample_shutter: ZebraShutter

    # SetEnergyComposite fields
    vfm: FocusingMirrorWithStripes
    mirror_voltages: MirrorVoltages
    dcm: DCM
    undulator_dcm: UndulatorDCM

    # RobotLoad fields
    robot: BartRobot
    webcam: Webcam
    lower_gonio: XYZPositioner
    beamstop: Beamstop


def create_devices(context: BlueskyContext) -> RobotLoadThenCentreComposite:
    from mx_bluesky.common.utils.context import device_composite_from_context

    return device_composite_from_context(context, RobotLoadThenCentreComposite)


def _flyscan_plan_from_robot_load_params(
    composite: RobotLoadThenCentreComposite,
    params: RobotLoadThenCentre,
    oav_config_file: str = OavConstants.OAV_CONFIG_JSON,
):
    yield from pin_centre_then_flyscan_plan(
        cast(GridDetectThenXRayCentreComposite, composite),
        params.pin_centre_then_xray_centre_params,
    )


def _robot_load_then_flyscan_plan(
    composite: RobotLoadThenCentreComposite,
    params: RobotLoadThenCentre,
    oav_config_file: str = OavConstants.OAV_CONFIG_JSON,
):
    yield from robot_load_and_change_energy_plan(
        cast(RobotLoadAndEnergyChangeComposite, composite),
        params.robot_load_params,
    )

    yield from _flyscan_plan_from_robot_load_params(composite, params, oav_config_file)


def robot_load_then_xray_centre(
    composite: RobotLoadThenCentreComposite,
    parameters: RobotLoadThenCentre,
) -> MsgGenerator:
    """Perform pin-tip detection followed by a flyscan to determine centres of interest.
    Performs a robot load if necessary."""
    eiger: EigerDetector = composite.eiger

    # TODO: get these from one source of truth #254
    assert parameters.sample_puck is not None
    assert parameters.sample_pin is not None

    sample_location = SampleLocation(parameters.sample_puck, parameters.sample_pin)

    doing_sample_load = not (
        yield from pin_already_loaded(composite.robot, sample_location)
    )

    current_chi = yield from bps.rd(composite.smargon.chi)
    LOGGER.info(f"Read back current smargon chi of {current_chi} degrees.")
    doing_chi_change = parameters.chi_start_deg is not None and not isclose(
        current_chi, parameters.chi_start_deg, abs_tol=0.001
    )

    if doing_sample_load:
        LOGGER.info("Pin not loaded, loading and centring")
        plan = _robot_load_then_flyscan_plan(
            composite,
            parameters,
        )
    else:
        # Robot load normally sets the energy so we should do this explicitly if no load is
        # being done
        demand_energy_ev = parameters.demand_energy_ev
        LOGGER.info(f"Setting the energy to {demand_energy_ev}eV")
        yield from set_energy_plan(
            demand_energy_ev, cast(SetEnergyComposite, composite)
        )

        if doing_chi_change:
            plan = _flyscan_plan_from_robot_load_params(composite, parameters)
            LOGGER.info("Pin already loaded but chi changed so centring")
        else:
            LOGGER.info("Pin already loaded and chi not changed so doing nothing")
            return

    detector_params = yield from fill_in_energy_if_not_supplied(
        composite.dcm, parameters.detector_params
    )

    eiger.set_detector_parameters(detector_params)

    yield from start_preparing_data_collection_then_do_plan(
        composite.beamstop,
        eiger,
        composite.detector_motion,
        parameters.detector_distance_mm,
        plan,
        group=CONST.WAIT.GRID_READY_FOR_DC,
    )



from __future__ import annotations

import os
from collections.abc import Callable, Sequence
from copy import deepcopy
from typing import Any, Literal

import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.synchrotron import SynchrotronMode
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    populate_data_collection_group,
    populate_remaining_data_collection_info,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_mapping import (
    construct_comment_for_gridscan,
    populate_xy_data_collection_info,
    populate_xz_data_collection_info,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    Orientation,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.parameters.components import IspybExperimentType
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    GridDetectThenXRayCentreComposite,
    grid_detect_then_xray_centre,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
    rotation_scan,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import (
    GridCommonWithHyperionDetectorParams,
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
)
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from ...conftest import (
    DATA_COLLECTION_COLUMN_MAP,
    compare_actual_and_expected,
    compare_comment,
)
from .conftest import raw_params_from_file
from .test_exp_eye_dev import SAMPLE_ID

EXPECTED_DATACOLLECTION_FOR_ROTATION = {
    "wavelength": 0.71,
    "beamSizeAtSampleX": 0.02,
    "beamSizeAtSampleY": 0.02,
    "exposureTime": 0.023,
    "undulatorGap1": 1.11,
    "synchrotronMode": SynchrotronMode.USER.value,
    "slitGapHorizontal": 0.123,
    "slitGapVertical": 0.234,
}

GRID_INFO_COLUMN_MAP = {
    s.lower(): s
    for s in [
        "gridInfoId",
        "dataCollectionGroupId",
        "xOffset",
        "yOffset",
        "dx_mm",
        "dy_mm",
        "steps_x",
        "steps_y",
        "meshAngle",
        "pixelsPerMicronX",
        "pixelsPerMicronY",
        "snapshot_offsetXPixel",
        "snapshot_offsetYPixel",
        "recordTimeStamp",
        "orientation",
        "workflowMeshId",
        "snaked",
        "dataCollectionId",
        "patchesX",
        "patchesY",
        "micronsPerPixelX",
        "micronsPerPixelY",
    ]
}


@pytest.fixture
def dummy_data_collection_group_info(dummy_params):
    return populate_data_collection_group(
        dummy_params,
    )


@pytest.fixture
def dummy_scan_data_info_for_begin_xy(dummy_params):
    info = populate_xy_data_collection_info(
        dummy_params.detector_params,
    )
    info = populate_remaining_data_collection_info(
        "MX-Bluesky: Xray centring 1 -", None, info, dummy_params
    )
    return ScanDataInfo(
        data_collection_info=info,
    )


@pytest.fixture
def dummy_scan_data_info_for_begin_xz(dummy_params):
    info = populate_xz_data_collection_info(
        dummy_params.detector_params,
    )
    info = populate_remaining_data_collection_info(
        "MX-Bluesky: Xray centring 2 -", None, info, dummy_params
    )
    return ScanDataInfo(
        data_collection_info=info,
    )


@pytest.fixture
def storage_directory(tmp_path) -> str:
    return str(tmp_path)


@pytest.fixture
def grid_detect_then_xray_centre_parameters(storage_directory):
    json_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/ispyb_gridscan_system_test_parameters.json"
    )
    json_dict["sample_id"] = SAMPLE_ID
    json_dict["visit"] = os.environ.get("ST_VISIT", "cm31105-4")
    json_dict["storage_directory"] = storage_directory
    return GridScanWithEdgeDetect(**json_dict)


def scan_xy_data_info_for_update(
    data_collection_group_id,
    dummy_params: HyperionSpecifiedThreeDGridScan,
    scan_data_info_for_begin,
):
    scan_data_info_for_update = deepcopy(scan_data_info_for_begin)
    scan_data_info_for_update.data_collection_info.parent_id = data_collection_group_id
    assert dummy_params is not None
    scan_data_info_for_update.data_collection_grid_info = DataCollectionGridInfo(
        dx_in_mm=dummy_params.x_step_size_um,
        dy_in_mm=dummy_params.y_step_size_um,
        steps_x=dummy_params.x_steps,
        steps_y=dummy_params.y_steps,
        microns_per_pixel_x=1.25,
        microns_per_pixel_y=1.25,
        # cast coordinates from numpy int64 to avoid mysql type conversion issues
        snapshot_offset_x_pixel=100,
        snapshot_offset_y_pixel=100,
        orientation=Orientation.HORIZONTAL,
        snaked=True,
    )
    scan_data_info_for_update.data_collection_info.comments = (
        construct_comment_for_gridscan(
            scan_data_info_for_update.data_collection_grid_info,
        )
    )
    return scan_data_info_for_update


def scan_data_infos_for_update_3d(
    ispyb_ids,
    scan_xy_data_info_for_update,
    dummy_params: HyperionSpecifiedThreeDGridScan,
):
    xz_data_collection_info = populate_xz_data_collection_info(
        dummy_params.detector_params
    )

    assert dummy_params is not None
    data_collection_grid_info = DataCollectionGridInfo(
        dx_in_mm=dummy_params.x_step_size_um,
        dy_in_mm=dummy_params.z_step_size_um,
        steps_x=dummy_params.x_steps,
        steps_y=dummy_params.z_steps,
        microns_per_pixel_x=1.25,
        microns_per_pixel_y=1.25,
        # cast coordinates from numpy int64 to avoid mysql type conversion issues
        snapshot_offset_x_pixel=100,
        snapshot_offset_y_pixel=50,
        orientation=Orientation.HORIZONTAL,
        snaked=True,
    )
    xz_data_collection_info = populate_remaining_data_collection_info(
        construct_comment_for_gridscan(data_collection_grid_info),
        ispyb_ids.data_collection_group_id,
        xz_data_collection_info,
        dummy_params,
    )
    xz_data_collection_info.parent_id = ispyb_ids.data_collection_group_id

    scan_xz_data_info_for_update = ScanDataInfo(
        data_collection_id=ispyb_ids.data_collection_ids[1],
        data_collection_info=xz_data_collection_info,
        data_collection_grid_info=(data_collection_grid_info),
    )
    return [scan_xy_data_info_for_update, scan_xz_data_info_for_update]


@pytest.mark.system_test
def test_ispyb_deposition_comment_correct_on_failure(
    dummy_ispyb: StoreInIspyb,
    fetch_comment: Callable[..., Any],
    dummy_data_collection_group_info,
    dummy_scan_data_info_for_begin_xy,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_data_collection_group_info, [dummy_scan_data_info_for_begin_xy]
    )
    dummy_ispyb.end_deposition(ispyb_ids, "fail", "could not connect to devices")
    assert (
        fetch_comment(ispyb_ids.data_collection_ids[0])  # type: ignore
        == "MX-Bluesky: Xray centring 1 - DataCollection Unsuccessful reason: could not connect to devices"
    )


@pytest.mark.system_test
def test_ispyb_deposition_comment_correct_for_3D_on_failure(
    dummy_ispyb_3d: StoreInIspyb,
    fetch_comment: Callable[..., Any],
    dummy_params,
    dummy_data_collection_group_info,
    dummy_scan_data_info_for_begin_xy,
    dummy_scan_data_info_for_begin_xz,
):
    ispyb_ids = dummy_ispyb_3d.begin_deposition(
        dummy_data_collection_group_info,
        [dummy_scan_data_info_for_begin_xy, dummy_scan_data_info_for_begin_xz],
    )
    scan_data_infos = generate_scan_data_infos(
        dummy_params,
        dummy_scan_data_info_for_begin_xy,
        IspybExperimentType.GRIDSCAN_3D,
        ispyb_ids,
    )
    ispyb_ids = dummy_ispyb_3d.update_deposition(ispyb_ids, scan_data_infos)
    dcid1 = ispyb_ids.data_collection_ids[0]  # type: ignore
    dcid2 = ispyb_ids.data_collection_ids[1]  # type: ignore
    dummy_ispyb_3d.end_deposition(ispyb_ids, "fail", "could not connect to devices")
    assert (
        fetch_comment(dcid1)
        == "MX-Bluesky: Xray centring 1 - Diffraction grid scan of 40 by 20 images in 100.0 um by 100.0 um steps. Top "
        "left (px): [100,100], bottom right (px): [3300,1700]. DataCollection Unsuccessful reason: could not connect to devices"
    )
    assert (
        fetch_comment(dcid2)
        == "MX-Bluesky: Xray centring 2 - Diffraction grid scan of 40 by 10 images in 100.0 um by 100.0 um steps. Top "
        "left (px): [100,50], bottom right (px): [3300,850]. DataCollection Unsuccessful reason: could not connect to devices"
    )


@pytest.mark.system_test
@pytest.mark.parametrize(
    "experiment_type, exp_num_of_grids, success",
    [
        (IspybExperimentType.GRIDSCAN_2D, 1, False),
        (IspybExperimentType.GRIDSCAN_2D, 1, True),
        (IspybExperimentType.GRIDSCAN_3D, 2, False),
        (IspybExperimentType.GRIDSCAN_3D, 2, True),
    ],
)
def test_can_store_2D_ispyb_data_correctly_when_in_error(
    experiment_type,
    exp_num_of_grids: Literal[1, 2],
    success: bool,
    fetch_comment: Callable[..., Any],
    dummy_params,
    dummy_data_collection_group_info,
    dummy_scan_data_info_for_begin_xy,
    dummy_scan_data_info_for_begin_xz,
):
    ispyb: StoreInIspyb = StoreInIspyb(CONST.SIM.DEV_ISPYB_DATABASE_CFG)
    scan_data_infos = [dummy_scan_data_info_for_begin_xy]
    if experiment_type == IspybExperimentType.GRIDSCAN_3D:
        scan_data_infos += [dummy_scan_data_info_for_begin_xz]
    ispyb_ids: IspybIds = ispyb.begin_deposition(
        dummy_data_collection_group_info, scan_data_infos
    )
    scan_data_infos = generate_scan_data_infos(
        dummy_params, dummy_scan_data_info_for_begin_xy, experiment_type, ispyb_ids
    )

    ispyb_ids = ispyb.update_deposition(ispyb_ids, scan_data_infos)
    assert len(ispyb_ids.data_collection_ids) == exp_num_of_grids  # type: ignore
    assert len(ispyb_ids.grid_ids) == exp_num_of_grids  # type: ignore
    assert isinstance(ispyb_ids.data_collection_group_id, int)

    expected_comments = [
        (
            "MX-Bluesky: Xray centring 1 - Diffraction grid scan of 40 by 20 "
            "images in 100.0 um by 100.0 um steps. Top left (px): [100,100], bottom right (px): [3300,1700]."
        ),
        (
            "MX-Bluesky: Xray centring 2 - Diffraction grid scan of 40 by 10 "
            "images in 100.0 um by 100.0 um steps. Top left (px): [100,50], bottom right (px): [3300,850]."
        ),
    ]

    if success:
        ispyb.end_deposition(ispyb_ids, "success", "")
    else:
        ispyb.end_deposition(ispyb_ids, "fail", "In error")
        expected_comments = [
            e + " DataCollection Unsuccessful reason: In error"
            for e in expected_comments
        ]

    assert (
        not isinstance(ispyb_ids.data_collection_ids, int)
        and ispyb_ids.data_collection_ids is not None
    )
    for grid_no, dc_id in enumerate(ispyb_ids.data_collection_ids):
        assert fetch_comment(dc_id) == expected_comments[grid_no]


@pytest.mark.system_test
def test_ispyb_deposition_in_gridscan(
    RE: RunEngine,
    grid_detect_then_xray_centre_composite: GridDetectThenXRayCentreComposite,
    grid_detect_then_xray_centre_parameters: GridScanWithEdgeDetect,
    fetch_datacollection_attribute: Callable[..., Any],
    fetch_datacollection_grid_attribute: Callable[..., Any],
    fetch_datacollection_position_attribute: Callable[..., Any],
    storage_directory: str,
):
    set_mock_value(
        grid_detect_then_xray_centre_composite.s4_slit_gaps.xgap.user_readback, 0.1
    )
    set_mock_value(
        grid_detect_then_xray_centre_composite.s4_slit_gaps.ygap.user_readback, 0.1
    )
    ispyb_callback = GridscanISPyBCallback(GridCommonWithHyperionDetectorParams)
    RE.subscribe(ispyb_callback)
    RE(
        grid_detect_then_xray_centre(
            grid_detect_then_xray_centre_composite,
            grid_detect_then_xray_centre_parameters,
        )
    )

    ispyb_ids = ispyb_callback.ispyb_ids
    DC_EXPECTED_VALUES = {
        "detectorid": 78,
        "axisstart": 0.0,
        "axisrange": 0,
        "axisend": 0,
        "focalspotsizeatsamplex": 0.02,
        "focalspotsizeatsampley": 0.02,
        "slitgapvertical": 0.1,
        "slitgaphorizontal": 0.1,
        "beamsizeatsamplex": 0.02,
        "beamsizeatsampley": 0.02,
        "transmission": 49.118,
        "datacollectionnumber": 1,
        "detectordistance": 100.0,
        "exposuretime": 0.12,
        "imagedirectory": f"{storage_directory}/",
        "imageprefix": "file_name",
        "imagesuffix": "h5",
        "numberofpasses": 1,
        "overlap": 0,
        "omegastart": 0,
        "startimagenumber": 1,
        "wavelength": 0.976254,
        "xbeam": 150.0,
        "ybeam": 160.0,
        "xtalsnapshotfullpath1": f"{storage_directory}/snapshots/file_name_1_0_grid_overlay.png",
        "xtalsnapshotfullpath2": f"{storage_directory}/snapshots/file_name_1_0_outer_overlay.png",
        "xtalsnapshotfullpath3": f"{storage_directory}/snapshots/file_name_1_0.png",
        "synchrotronmode": "User",
        "undulatorgap1": 1.11,
        "filetemplate": "file_name_1_master.h5",
        "numberofimages": 20 * 6,
    }
    compare_comment(
        fetch_datacollection_attribute,
        ispyb_ids.data_collection_ids[0],
        "MX-Bluesky: Xray centring 1 - Diffraction grid scan of 20 by 6 "
        "images in 20.0 um by 20.0 um steps. Top left (px): [130,130], "
        "bottom right (px): [626,278]. Aperture: Small. ",
    )
    compare_actual_and_expected(
        ispyb_ids.data_collection_ids[0],
        DC_EXPECTED_VALUES,
        fetch_datacollection_attribute,
        DATA_COLLECTION_COLUMN_MAP,
    )
    GRIDINFO_EXPECTED_VALUES = {
        "gridInfoId": ispyb_ids.grid_ids[0],
        "dx_mm": 0.02,
        "dy_mm": 0.02,
        "steps_x": 20,
        "steps_y": 6,
        "snapshot_offsetXPixel": 130,
        "snapshot_offsetYPixel": 130,
        "orientation": "horizontal",
        "snaked": True,
        "dataCollectionId": ispyb_ids.data_collection_ids[0],
        "micronsPerPixelX": 0.806,
        "micronsPerPixelY": 0.806,
    }

    compare_actual_and_expected(
        ispyb_ids.grid_ids[0],
        GRIDINFO_EXPECTED_VALUES,
        fetch_datacollection_grid_attribute,
        GRID_INFO_COLUMN_MAP,
    )
    position_id = fetch_datacollection_attribute(
        ispyb_ids.data_collection_ids[0], DATA_COLLECTION_COLUMN_MAP["positionid"]
    )
    assert position_id is None
    DC_EXPECTED_VALUES.update(
        {
            "axisstart": 90.0,
            "axisend": 90.0,
            "datacollectionnumber": 2,
            "omegastart": 90.0,
            "filetemplate": "file_name_2_master.h5",
            "xtalsnapshotfullpath1": f"{storage_directory}/snapshots/file_name_1_90_grid_overlay.png",
            "xtalsnapshotfullpath2": f"{storage_directory}/snapshots/file_name_1_90_outer_overlay.png",
            "xtalsnapshotfullpath3": f"{storage_directory}/snapshots/file_name_1_90.png",
            "numberofimages": 20 * 6,
        }
    )
    compare_actual_and_expected(
        ispyb_ids.data_collection_ids[1],
        DC_EXPECTED_VALUES,
        fetch_datacollection_attribute,
        DATA_COLLECTION_COLUMN_MAP,
    )
    compare_comment(
        fetch_datacollection_attribute,
        ispyb_ids.data_collection_ids[1],
        "MX-Bluesky: Xray centring 2 - Diffraction grid scan of 20 by 6 "
        "images in 20.0 um by 20.0 um steps. Top left (px): [130,130], "
        "bottom right (px): [626,278]. Aperture: Small. ",
    )
    position_id = fetch_datacollection_attribute(
        ispyb_ids.data_collection_ids[1], DATA_COLLECTION_COLUMN_MAP["positionid"]
    )
    assert position_id is None
    GRIDINFO_EXPECTED_VALUES.update(
        {
            "gridInfoId": ispyb_ids.grid_ids[1],
            "steps_y": 6.0,
            "snapshot_offsetYPixel": 130.0,
            "dataCollectionId": ispyb_ids.data_collection_ids[1],
        }
    )
    compare_actual_and_expected(
        ispyb_ids.grid_ids[1],
        GRIDINFO_EXPECTED_VALUES,
        fetch_datacollection_grid_attribute,
        GRID_INFO_COLUMN_MAP,
    )


@pytest.mark.system_test
def test_ispyb_deposition_in_rotation_plan(
    composite_for_rotation_scan: RotationScanComposite,
    params_for_rotation_scan: RotationScan,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    fetch_comment: Callable[..., Any],
    fetch_datacollection_attribute: Callable[..., Any],
    fetch_datacollection_position_attribute: Callable[..., Any],
    feature_flags_update_with_omega_flip,
):
    ispyb_cb = RotationISPyBCallback()
    RE.subscribe(ispyb_cb)

    RE(
        rotation_scan(
            composite_for_rotation_scan,
            params_for_rotation_scan,
            oav_parameters_for_rotation,
        )
    )

    dcid = ispyb_cb.ispyb_ids.data_collection_ids[0]
    assert dcid is not None
    assert (
        fetch_comment(dcid) == "test Sample position (µm): (1, 2, 3) Aperture: Small. "
    )

    expected_values = EXPECTED_DATACOLLECTION_FOR_ROTATION | {
        "xtalSnapshotFullPath1": "regex:/tmp/dls/i03/data/2024/cm31105-4/auto/123456/snapshots/\\d{8}_oav_snapshot_0"
        ".png",
        "xtalSnapshotFullPath2": "regex:/tmp/dls/i03/data/2024/cm31105-4/auto/123456/snapshots/\\d{8}_oav_snapshot_90"
        ".png",
        "xtalSnapshotFullPath3": "regex:/tmp/dls/i03/data/2024/cm31105-4/auto/123456/snapshots/\\d{8}_oav_snapshot_180"
        ".png",
        "xtalSnapshotFullPath4": "regex:/tmp/dls/i03/data/2024/cm31105-4/auto/123456/snapshots/\\d{8}_oav_snapshot_270"
        ".png",
    }

    compare_actual_and_expected(dcid, expected_values, fetch_datacollection_attribute)

    position_id = fetch_datacollection_attribute(
        dcid, DATA_COLLECTION_COLUMN_MAP["positionid"]
    )
    expected_values = {"posX": 0.001, "posY": 0.002, "posZ": 0.003}
    compare_actual_and_expected(
        position_id, expected_values, fetch_datacollection_position_attribute
    )


def generate_scan_data_infos(
    dummy_params,
    dummy_scan_data_info_for_begin: ScanDataInfo,
    experiment_type: IspybExperimentType,
    ispyb_ids: IspybIds,
) -> Sequence[ScanDataInfo]:
    xy_scan_data_info = scan_xy_data_info_for_update(
        ispyb_ids.data_collection_group_id, dummy_params, dummy_scan_data_info_for_begin
    )
    xy_scan_data_info.data_collection_id = ispyb_ids.data_collection_ids[0]
    if experiment_type == IspybExperimentType.GRIDSCAN_3D:
        scan_data_infos = scan_data_infos_for_update_3d(
            ispyb_ids, xy_scan_data_info, dummy_params
        )
    else:
        scan_data_infos = [xy_scan_data_info]
    return scan_data_infos



from collections.abc import Callable
from typing import Any
from unittest.mock import MagicMock

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pytest
from bluesky.run_engine import RunEngine
from ophyd_async.sim import SimMotor

from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGroupInfo,
)


class MockReactiveCallback(PlanReactiveCallback):
    activity_gated_start: MagicMock
    activity_gated_descriptor: MagicMock
    activity_gated_event: MagicMock
    activity_gated_stop: MagicMock

    def __init__(self, *, emit: Callable[..., Any] | None = None) -> None:
        super().__init__(MagicMock(), emit=emit)
        self.activity_gated_start = MagicMock(name="activity_gated_start")  # type: ignore
        self.activity_gated_descriptor = MagicMock(name="activity_gated_descriptor")  # type: ignore
        self.activity_gated_event = MagicMock(name="activity_gated_event")  # type: ignore
        self.activity_gated_stop = MagicMock(name="activity_gated_stop")  # type: ignore


@pytest.fixture
def mocked_test_callback():
    t = MockReactiveCallback()
    return t


@pytest.fixture
def RE_with_mock_callback(mocked_test_callback):
    RE = RunEngine()
    RE.subscribe(mocked_test_callback)
    yield RE, mocked_test_callback


def get_test_plan(callback_name):
    s = SimMotor(name="fake_signal")

    @bpp.run_decorator(md={"activate_callbacks": [callback_name]})
    def test_plan():
        yield from bps.create()
        yield from bps.read(s)
        yield from bps.save()

    return test_plan, s


@pytest.fixture
def dummy_rotation_data_collection_group_info():
    return DataCollectionGroupInfo(
        visit_string="cm31105-4",
        experiment_type="SAD",
        sample_id=364758,
    )



from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, cast

from mx_bluesky.common.external_interaction.callbacks.common.ispyb_callback_base import (
    BaseISPyBCallback,
)
from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    populate_data_collection_group,
    populate_remaining_data_collection_info,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionInfo,
    DataCollectionPositionInfo,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.parameters.components import IspybExperimentType
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER, set_dcgid_tag
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_mapping import (
    populate_data_collection_info_for_rotation,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import RotationScan

if TYPE_CHECKING:
    from event_model.documents import Event, RunStart, RunStop


class RotationISPyBCallback(BaseISPyBCallback):
    """Callback class to handle the deposition of experiment parameters into the ISPyB
    database. Listens for 'event' and 'descriptor' documents. Creates the ISpyB entry on
    recieving an 'event' document for the 'ispyb_reading_hardware' event, and updates the
    deposition on recieving its final 'stop' document.

    To use, subscribe the Bluesky RunEngine to an instance of this class.
    E.g.:
        ispyb_handler_callback = RotationISPyBCallback(parameters)
        RE.subscribe(ispyb_handler_callback)
    Or decorate a plan using bluesky.preprocessors.subs_decorator.

    See: https://blueskyproject.io/bluesky/callbacks.html#ways-to-invoke-callbacks
    """

    def __init__(
        self,
        *,
        emit: Callable[..., Any] | None = None,
    ) -> None:
        super().__init__(emit=emit)
        self.last_sample_id: int | None = None
        self.ispyb_ids: IspybIds = IspybIds()

    def activity_gated_start(self, doc: RunStart):
        if doc.get("subplan_name") == CONST.PLAN.ROTATION_OUTER:
            ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                "ISPyB callback received start document with experiment parameters."
            )
            hyperion_params = doc.get("mx_bluesky_parameters")
            assert isinstance(hyperion_params, str)
            self.params = RotationScan.model_validate_json(hyperion_params)
            dcgid = (
                self.ispyb_ids.data_collection_group_id
                if (self.params.sample_id == self.last_sample_id)
                else None
            )
            if (
                self.params.ispyb_experiment_type
                == IspybExperimentType.CHARACTERIZATION
            ):
                ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                    "Screening collection - using new DCG"
                )
                dcgid = None
                self.last_sample_id = None
            else:
                ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                    f"Collection is {self.params.ispyb_experiment_type} - storing sampleID to bundle images"
                )
                self.last_sample_id = self.params.sample_id
            self.ispyb = StoreInIspyb(self.ispyb_config)
            ISPYB_ZOCALO_CALLBACK_LOGGER.info("Beginning ispyb deposition")
            data_collection_group_info = populate_data_collection_group(self.params)
            data_collection_info = populate_data_collection_info_for_rotation(
                cast(RotationScan, self.params)
            )
            data_collection_info = populate_remaining_data_collection_info(
                self.params.comment,
                dcgid,
                data_collection_info,
                self.params,
            )
            data_collection_info.parent_id = dcgid
            scan_data_info = ScanDataInfo(
                data_collection_info=data_collection_info,
            )
            self.ispyb_ids = self.ispyb.begin_deposition(
                data_collection_group_info, [scan_data_info]
            )
        ISPYB_ZOCALO_CALLBACK_LOGGER.info("ISPYB handler received start document.")
        if doc.get("subplan_name") == CONST.PLAN.ROTATION_MAIN:
            self.uid_to_finalize_on = doc.get("uid")
        return super().activity_gated_start(doc)

    def populate_info_for_update(
        self,
        event_sourced_data_collection_info: DataCollectionInfo,
        event_sourced_position_info: DataCollectionPositionInfo | None,
        params,
    ) -> Sequence[ScanDataInfo]:
        assert self.ispyb_ids.data_collection_ids, (
            "Expect an existing DataCollection to update"
        )

        return [
            ScanDataInfo(
                data_collection_info=event_sourced_data_collection_info,
                data_collection_id=self.ispyb_ids.data_collection_ids[0],
                data_collection_position_info=event_sourced_position_info,
            )
        ]

    def _handle_ispyb_hardware_read(self, doc: Event):
        """Use the hardware read values to create the ispyb comment"""
        scan_data_infos = super()._handle_ispyb_hardware_read(doc)
        motor_positions_mm = [
            doc["data"]["smargon-x"],
            doc["data"]["smargon-y"],
            doc["data"]["smargon-z"],
        ]
        assert self.params, (
            "handle_ispyb_hardware_read triggered before activity_gated_start"
        )
        motor_positions_um = [position * 1000 for position in motor_positions_mm]
        comment = f"Sample position (µm): ({motor_positions_um[0]:.0f}, {motor_positions_um[1]:.0f}, {motor_positions_um[2]:.0f})"
        scan_data_infos[0].data_collection_info.comments = comment
        return scan_data_infos

    def activity_gated_event(self, doc: Event):
        doc = super().activity_gated_event(doc)
        set_dcgid_tag(self.ispyb_ids.data_collection_group_id)

        descriptor_name = self.descriptors[doc["descriptor"]].get("name")
        if descriptor_name == CONST.DESCRIPTORS.OAV_ROTATION_SNAPSHOT_TRIGGERED:
            scan_data_infos = self._handle_oav_rotation_snapshot_triggered(doc)
            self.ispyb_ids = self.ispyb.update_deposition(
                self.ispyb_ids, scan_data_infos
            )

        return doc

    def _handle_oav_rotation_snapshot_triggered(self, doc) -> Sequence[ScanDataInfo]:
        assert self.ispyb_ids.data_collection_ids, "No current data collection"
        assert self.params, "ISPyB handler didn't receive parameters!"
        data = doc["data"]
        self._oav_snapshot_event_idx += 1
        data_collection_info = DataCollectionInfo(
            **{
                f"xtal_snapshot{self._oav_snapshot_event_idx}": data.get(
                    "oav-snapshot-last_saved_path"
                )
            }
        )
        scan_data_info = ScanDataInfo(
            data_collection_id=self.ispyb_ids.data_collection_ids[-1],
            data_collection_info=data_collection_info,
        )
        return [scan_data_info]

    def activity_gated_stop(self, doc: RunStop) -> RunStop:
        if doc.get("run_start") == self.uid_to_finalize_on:
            self.uid_to_finalize_on = None
            return super().activity_gated_stop(doc)
        return self._tag_doc(doc)



from bluesky import plan_stubs as bps
from dodal.devices.detector.detector_motion import DetectorMotion, ShutterState

from mx_bluesky.common.utils.log import LOGGER


def set_detector_z_position(
    detector_motion: DetectorMotion, detector_position: float, group=None
):
    LOGGER.info(f"Moving detector to {detector_position} ({group})")
    yield from bps.abs_set(detector_motion.z, detector_position, group=group)


def set_shutter(detector_motion: DetectorMotion, state: ShutterState, group=None):
    LOGGER.info(f"Setting shutter to {state} ({group})")
    yield from bps.abs_set(detector_motion.shutter, state, group=group)



"""
Move on click gui for fixed targets at I24
Robin Owen 12 Jan 2021
"""

from collections.abc import Sequence

import bluesky.plan_stubs as bps
import cv2 as cv
from bluesky.run_engine import RunEngine
from dodal.beamlines import i24
from dodal.devices.i24.pmac import PMAC
from dodal.devices.oav.oav_detector import OAV

from mx_bluesky.beamlines.i24.serial.fixed_target import (
    i24ssx_Chip_Manager_py3v1 as manager,
)
from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import Fiducials
from mx_bluesky.beamlines.i24.serial.log import SSX_LOGGER
from mx_bluesky.beamlines.i24.serial.parameters.constants import OAV1_CAM


def _get_beam_centre(oav: OAV):
    """Extract the beam centre x/y positions from the display.configuration file.

    Args:
        oav (OAV): the OAV device.
    """
    beam_x = yield from bps.rd(oav.beam_centre_i)
    beam_y = yield from bps.rd(oav.beam_centre_j)
    return beam_x, beam_y


def _calculate_zoom_calibrator(oav: OAV):
    """Set the scale for the zoom calibrator for the pmac moves."""
    currentzoom = yield from bps.rd(oav.zoom_controller.percentage)
    zoomcalibrator = (
        1.285
        - (0.02866 * currentzoom)
        + (0.00025 * currentzoom**2)
        - (0.0000008151 * currentzoom**3)
    )
    return zoomcalibrator


def _move_on_mouse_click_plan(
    oav: OAV,
    pmac: PMAC,
    clicked_position: Sequence[int],
):
    """A plan that calculates the zoom calibrator and moves to the clicked \
        position coordinates.
    """
    zoomcalibrator = yield from _calculate_zoom_calibrator(oav)
    beamX, beamY = yield from _get_beam_centre(oav)
    x, y = clicked_position
    xmove = -10 * (beamX - x) * zoomcalibrator
    ymove = 10 * (beamY - y) * zoomcalibrator
    SSX_LOGGER.info(f"Zoom calibrator {zoomcalibrator}")
    SSX_LOGGER.info(f"Beam centre {beamX} {beamY}")
    SSX_LOGGER.info(f"Moving X and Y {xmove} {ymove}")
    xmovepmacstring = "#1J:" + str(xmove)
    ymovepmacstring = "#2J:" + str(ymove)
    yield from bps.abs_set(pmac.pmac_string, xmovepmacstring, wait=True)
    yield from bps.abs_set(pmac.pmac_string, ymovepmacstring, wait=True)


# Register clicks and move chip stages
def onMouse(event, x, y, flags, param):
    if event == cv.EVENT_LBUTTONUP:
        RE = param[0]
        pmac = param[1]
        oav = param[2]
        SSX_LOGGER.info(f"Clicked X and Y {x} {y}")
        RE(_move_on_mouse_click_plan(oav, pmac, (x, y)))


def update_ui(oav, frame, RE):
    # Get beam x and y values
    beamX, beamY = RE(_get_beam_centre(oav)).plan_result

    # Overlay text and beam centre
    cv.ellipse(
        frame, (beamX, beamY), (12, 8), 0.0, 0.0, 360, (0, 255, 255), thickness=2
    )
    cv.putText(
        frame,
        "Key bindings",
        (20, 40),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        1,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "Q / A : go to / set as f0",
        (25, 70),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "W / S : go to / set as f1",
        (25, 90),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "E / D : go to / set as f2",
        (25, 110),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "I / O : in /out of focus",
        (25, 130),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "C : Create CS",
        (25, 150),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "esc : close window",
        (25, 170),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.imshow("OAV1view", frame)


def start_viewer(oav: OAV, pmac: PMAC, RE: RunEngine, oav1: str = OAV1_CAM):
    # Create a video caputure from OAV1
    cap = cv.VideoCapture(oav1)

    # Create window named OAV1view and set onmouse to this
    cv.namedWindow("OAV1view")
    cv.setMouseCallback("OAV1view", onMouse, param=[RE, pmac, oav])  # type: ignore

    SSX_LOGGER.info("Showing camera feed. Press escape to close")
    # Read captured video and store them in success and frame
    success, frame = cap.read()

    # Loop until escape key is pressed. Keyboard shortcuts here
    while success:
        success, frame = cap.read()

        update_ui(oav, frame, RE)

        k = cv.waitKey(1)
        if k == 113:  # Q
            RE(manager.moveto(Fiducials.zero, pmac))
        if k == 119:  # W
            RE(manager.moveto(Fiducials.fid1, pmac))
        if k == 101:  # E
            RE(manager.moveto(Fiducials.fid2, pmac))
        if k == 97:  # A
            RE(bps.trigger(pmac.home, wait=True))
            print("Current position set as origin")
        if k == 115:  # S
            RE(manager.fiducial(1))
        if k == 100:  # D
            RE(manager.fiducial(2))
        if k == 99:  # C
            RE(manager.cs_maker(pmac))
        if k == 98:  # B
            RE(
                manager.block_check()
            )  # doesn't work well for blockcheck as image doesn't update
        if k == 104:  # H
            RE(bps.abs_set(pmac.pmac_string, "#2J:-10", wait=True))
        if k == 110:  # N
            RE(bps.abs_set(pmac.pmac_string, "#2J:10", wait=True))
        if k == 109:  # M
            RE(bps.abs_set(pmac.pmac_string, "#1J:-10", wait=True))
        if k == 98:  # B
            RE(bps.abs_set(pmac.pmac_string, "#1J:10", wait=True))
        if k == 105:  # I
            RE(bps.abs_set(pmac.pmac_string, "#3J:-150", wait=True))
        if k == 111:  # O
            RE(bps.abs_set(pmac.pmac_string, "#3J:150", wait=True))
        if k == 117:  # U
            RE(bps.abs_set(pmac.pmac_string, "#3J:-1000", wait=True))
        if k == 112:  # P
            RE(bps.abs_set(pmac.pmac_string, "#3J:1000", wait=True))
        if k == 0x1B:  # esc
            cv.destroyWindow("OAV1view")
            print("Pressed escape. Closing window")
            break

    # Clear cameraCapture instance
    cap.release()


if __name__ == "__main__":
    RE = RunEngine(call_returns_result=True)
    # Get devices out of dodal
    oav: OAV = i24.oav(connect_immediately=True)
    pmac: PMAC = i24.pmac(connect_immediately=True)
    start_viewer(oav, pmac, RE)



from bluesky import plan_stubs as bps
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.xbpm_feedback import Pause, XBPMFeedback

from mx_bluesky.common.utils.log import LOGGER


def unpause_xbpm_feedback_and_set_transmission_to_1(
    xbpm_feedback: XBPMFeedback, attenuator: BinaryFilterAttenuator
):
    """Turns the XBPM feedback back on and sets transmission to 1 so that it keeps the
    beam aligned whilst not collecting.

    Args:
        xbpm_feedback (XBPMFeedback): The XBPM device that is responsible for keeping
                                      the beam in position
        attenuator (BinaryFilterAttenuator): The attenuator used to set transmission
    """
    yield from bps.mv(xbpm_feedback.pause_feedback, Pause.RUN, attenuator, 1.0)  # type: ignore # See: https://github.com/bluesky/bluesky/issues/1809


def check_and_pause_feedback(
    xbpm_feedback: XBPMFeedback,
    attenuator: BinaryFilterAttenuator,
    desired_transmission_fraction: float,
):
    """Checks that the xbpm is in position before then turning it off and setting a new
    transmission.

    Args:
        xbpm_feedback (XBPMFeedback): The XBPM device that is responsible for keeping
                                      the beam in position
        attenuator (BinaryFilterAttenuator): The attenuator used to set transmission
        desired_transmission_fraction (float): The desired transmission to set after
                                               turning XBPM feedback off.

    """
    yield from bps.mv(attenuator, 1.0)  # type: ignore # See: https://github.com/bluesky/bluesky/issues/1809
    LOGGER.info("Waiting for XBPM feedback to be stable")
    yield from bps.trigger(xbpm_feedback, wait=True)
    LOGGER.info(
        f"XPBM feedback in position, pausing and setting transmission to {desired_transmission_fraction}"
    )
    yield from bps.mv(xbpm_feedback.pause_feedback, Pause.PAUSE)  # type: ignore # See: https://github.com/bluesky/bluesky/issues/1809
    yield from bps.mv(attenuator, desired_transmission_fraction)  # type: ignore # See: https://github.com/bluesky/bluesky/issues/1809



from unittest.mock import ANY, patch

import pytest
from requests import JSONDecodeError

from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import (
    BearerAuth,
    BLSampleStatus,
    ExpeyeInteraction,
    _get_base_url_and_token,
)
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMade


def test_get_url_and_token_returns_expected_data():
    url, token = _get_base_url_and_token()
    assert url == "http://blah"
    assert token == "notatoken"


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.post")
def test_when_start_load_called_then_correct_expected_url_posted_to_with_expected_data(
    mock_post,
):
    expeye_interactor = ExpeyeInteraction()
    expeye_interactor.start_load("test", 3, 700, 10, 5)

    mock_post.assert_called_once()
    assert (
        mock_post.call_args.args[0]
        == "http://blah/proposals/test/sessions/3/robot-actions"
    )
    expected_data = {
        "startTimestamp": ANY,
        "sampleId": 700,
        "actionType": "LOAD",
        "containerLocation": 5,
        "dewarLocation": 10,
    }
    assert mock_post.call_args.kwargs["json"] == expected_data


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.post")
def test_when_start_called_then_returns_id(mock_post):
    mock_post.return_value.json.return_value = {"robotActionId": 190}
    expeye_interactor = ExpeyeInteraction()
    robot_id = expeye_interactor.start_load("test", 3, 700, 10, 5)
    assert robot_id == 190


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.post")
def test_when_bad_response_no_json_handled_correctly(mock_post):
    mock_post.return_value.ok = False
    mock_post.return_value.json.side_effect = JSONDecodeError("Unable to decode", "", 0)

    expeye_interactor = ExpeyeInteraction()
    with pytest.raises(ISPyBDepositionNotMade):
        expeye_interactor.start_load("test", 3, 700, 10, 5)


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.post")
def test_when_start_load_called_then_use_correct_token(
    mock_post,
):
    expeye_interactor = ExpeyeInteraction()
    expeye_interactor.start_load("test", 3, 700, 10, 5)

    assert isinstance(auth := mock_post.call_args.kwargs["auth"], BearerAuth)
    assert auth.token == "notatoken"


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.post")
def test_given_server_does_not_respond_when_start_load_called_then_error(mock_post):
    mock_post.return_value.ok = False

    expeye_interactor = ExpeyeInteraction()
    with pytest.raises(ISPyBDepositionNotMade):
        expeye_interactor.start_load("test", 3, 700, 10, 5)


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.patch")
def test_when_end_load_called_with_success_then_correct_expected_url_posted_to_with_expected_data(
    # mocks HTTP PATCH
    mock_patch,
):
    expeye_interactor = ExpeyeInteraction()
    expeye_interactor.end_load(3, "success", "")

    mock_patch.assert_called_once()
    assert mock_patch.call_args.args[0] == "http://blah/robot-actions/3"
    expected_data = {
        "endTimestamp": ANY,
        "status": "SUCCESS",
        "message": "",
    }
    assert mock_patch.call_args.kwargs["json"] == expected_data


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.patch")
def test_when_end_load_called_with_failure_then_correct_expected_url_posted_to_with_expected_data(
    mock_patch,
):
    expeye_interactor = ExpeyeInteraction()
    expeye_interactor.end_load(3, "fail", "bad")

    mock_patch.assert_called_once()
    assert mock_patch.call_args.args[0] == "http://blah/robot-actions/3"
    expected_data = {
        "endTimestamp": ANY,
        "status": "ERROR",
        "message": "bad",
    }
    assert mock_patch.call_args.kwargs["json"] == expected_data


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.patch")
def test_when_end_load_called_then_use_correct_token(
    mock_patch,
):
    expeye_interactor = ExpeyeInteraction()
    expeye_interactor.end_load(3, "success", "")

    assert isinstance(auth := mock_patch.call_args.kwargs["auth"], BearerAuth)
    assert auth.token == "notatoken"


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.patch")
def test_given_server_does_not_respond_when_end_load_called_then_error(mock_patch):
    mock_patch.return_value.ok = False

    expeye_interactor = ExpeyeInteraction()
    with pytest.raises(ISPyBDepositionNotMade):
        expeye_interactor.end_load(1, "", "")


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.patch")
def test_when_update_barcode_called_with_success_then_correct_expected_url_posted_to_with_expected_data(
    mock_patch,
):
    expeye_interactor = ExpeyeInteraction()
    expeye_interactor.update_barcode_and_snapshots(
        3, "test", "/tmp/before.jpg", "/tmp/after.jpg"
    )

    mock_patch.assert_called_once()
    assert mock_patch.call_args.args[0] == "http://blah/robot-actions/3"
    expected_data = {
        "sampleBarcode": "test",
        "xtalSnapshotBefore": "/tmp/before.jpg",
        "xtalSnapshotAfter": "/tmp/after.jpg",
    }
    assert mock_patch.call_args.kwargs["json"] == expected_data


@patch("mx_bluesky.common.external_interaction.ispyb.exp_eye_store.patch")
def test_update_sample_status(
    mock_patch,
):
    expeye = ExpeyeInteraction()
    expected_json = {"blSampleStatus": "LOADED"}
    expeye.update_sample_status(12345, BLSampleStatus.LOADED)
    mock_patch.assert_called_with(
        "http://blah/samples/12345", auth=ANY, json=expected_json
    )



from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.backlight import BacklightPosition
from dodal.devices.detector.detector_motion import ShutterState
from dodal.devices.i03.beamstop import BeamstopPositions
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import SynchrotronMode

from mx_bluesky.hyperion.device_setup_plans.check_beamstop import BeamstopException
from mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan import (
    _fire_xray_centre_result_event,
)
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan import (
    create_parameters_for_grid_detection,
    pin_centre_then_flyscan_plan,
    pin_tip_centre_then_xray_centre,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import (
    PinTipCentreThenXrayCentre,
)

from ....conftest import TEST_RESULT_LARGE, raw_params_from_file, simulate_xrc_result
from .conftest import FLYSCAN_RESULT_LOW, FLYSCAN_RESULT_MED, sim_fire_event_on_open_run


@pytest.fixture
def test_pin_centre_then_xray_centre_params():
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_pin_centre_then_xray_centre_parameters.json"
    )
    return PinTipCentreThenXrayCentre(**params)


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.change_aperture_then_move_to_xtal",
    autospec=True,
)
def test_pin_tip_centre_then_xray_centre_moves_to_centre_of_first_flyscan_result(
    mock_change_aperture_then_move_to_xtal: MagicMock,
    mock_detect_and_do_gridscan: MagicMock,
    mock_pin_tip_centre: MagicMock,
    grid_detect_devices: GridDetectThenXRayCentreComposite,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
    test_config_files,
):
    mock_detect_and_do_gridscan.side_effect = lambda *_: _fire_xray_centre_result_event(
        [FLYSCAN_RESULT_MED, FLYSCAN_RESULT_LOW]
    )
    RE = RunEngine()
    RE(
        pin_tip_centre_then_xray_centre(
            grid_detect_devices,
            test_pin_centre_then_xray_centre_params,
            test_config_files["oav_config_json"],
        )
    )

    mock_detect_and_do_gridscan.assert_called_once()
    mock_pin_tip_centre.assert_called_once()
    mock_change_aperture_then_move_to_xtal.assert_called_once()
    assert (
        mock_change_aperture_then_move_to_xtal.mock_calls[0].args[0]
        == FLYSCAN_RESULT_MED
    )


def test_when_create_parameters_for_grid_detection_then_parameters_created(
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
):
    grid_detect_params = create_parameters_for_grid_detection(
        test_pin_centre_then_xray_centre_params
    )

    assert grid_detect_params.exposure_time_s == 0.1


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
def test_when_pin_centre_xray_centre_called_then_plan_runs_correctly(
    mock_detect_and_do_gridscan: MagicMock,
    mock_pin_tip_centre: MagicMock,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
    grid_detect_devices: GridDetectThenXRayCentreComposite,
    test_config_files,
):
    RE = RunEngine()
    RE(
        pin_centre_then_flyscan_plan(
            grid_detect_devices,
            test_pin_centre_then_xray_centre_params,
            test_config_files["oav_config_json"],
        )
    )

    mock_detect_and_do_gridscan.assert_called_once()
    mock_pin_tip_centre.assert_called_once()


@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.GridDetectionCallback",
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.grid_detection_plan",
    autospec=True,
)
def test_when_pin_centre_xray_centre_called_then_detector_positioned(
    mock_grid_detect: MagicMock,
    mock_pin_tip_centre: MagicMock,
    mock_grid_callback: MagicMock,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
    grid_detect_devices: GridDetectThenXRayCentreComposite,
    test_config_files,
    sim_run_engine: RunEngineSimulator,
):
    mock_grid_callback.return_value.get_grid_parameters.return_value = {
        "transmission_frac": 1.0,
        "exposure_time_s": 0,
        "x_start_um": 0,
        "y_start_um": 0,
        "y2_start_um": 0,
        "z_start_um": 0,
        "z2_start_um": 0,
        "x_steps": 10,
        "y_steps": 10,
        "z_steps": 10,
        "x_step_size_um": 0.1,
        "y_step_size_um": 0.1,
        "z_step_size_um": 0.1,
    }

    sim_run_engine.add_handler_for_callback_subscribes()

    sim_run_engine.add_handler(
        "read",
        lambda msg_: {"values": {"value": SynchrotronMode.SHUTDOWN}},
        "synchrotron-synchrotron_mode",
    )
    simulate_xrc_result(sim_run_engine, grid_detect_devices.zocalo, TEST_RESULT_LARGE)

    def add_handlers_to_simulate_detector_motion(msg: Msg):
        sim_run_engine.add_handler(
            "read",
            lambda msg_: {"values": {"value": int(ShutterState.OPEN)}},
            "detector_motion_shutter",
        )
        sim_run_engine.add_handler(
            "read",
            lambda msg_: {"values": {"value": 1}},
            "detector_motion_z_motor_done_move",
        )

    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    sim_run_engine.add_wait_handler(
        add_handlers_to_simulate_detector_motion, CONST.WAIT.GRID_READY_FOR_DC
    )

    simulate_xrc_result(sim_run_engine, grid_detect_devices.zocalo, TEST_RESULT_LARGE)
    messages = sim_run_engine.simulate_plan(
        pin_tip_centre_then_xray_centre(
            grid_detect_devices,
            test_pin_centre_then_xray_centre_params,
            test_config_files["oav_config_json"],
        ),
    )

    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.obj is grid_detect_devices.detector_motion.z
    )
    assert messages[0].args[0] == 100
    assert messages[0].kwargs["group"] == CONST.WAIT.GRID_READY_FOR_DC
    assert messages[1].obj is grid_detect_devices.detector_motion.shutter
    assert messages[1].args[0] == ShutterState.OPEN
    assert messages[1].kwargs["group"] == CONST.WAIT.GRID_READY_FOR_DC
    messages = assert_message_and_return_remaining(
        messages[2:],
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == CONST.WAIT.GRID_READY_FOR_DC,
    )
    assert_message_and_return_remaining(
        messages[2:],
        lambda msg: msg.command == "open_run"
        and msg.kwargs["subplan_name"] == "do_fgs",
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
def test_pin_centre_then_xray_centre_plan_activates_ispyb_callback_before_pin_tip_centre_plan(
    mock_detect_grid_and_do_gridscan,
    mock_pin_tip_centre_plan,
    sim_run_engine: RunEngineSimulator,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
    grid_detect_devices: GridDetectThenXRayCentreComposite,
    test_config_files,
):
    mock_detect_grid_and_do_gridscan.return_value = iter(
        [Msg("detect_grid_and_do_gridscan")]
    )
    mock_pin_tip_centre_plan.return_value = iter([Msg("pin_tip_centre_plan")])

    msgs = sim_run_engine.simulate_plan(
        pin_centre_then_flyscan_plan(
            grid_detect_devices,
            test_pin_centre_then_xray_centre_params,
            test_config_files["oav_config_json"],
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run"
        and "GridscanISPyBCallback" in msg.kwargs["activate_callbacks"],
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "pin_tip_centre_plan"
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "detect_grid_and_do_gridscan"
    )
    assert_message_and_return_remaining(msgs, lambda msg: msg.command == "close_run")


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
def test_pin_centre_then_xray_centre_plan_sets_up_backlight_and_aperture(
    mock_detect_grid_and_do_gridscan,
    mock_pin_tip_centre_plan,
    grid_detect_devices: GridDetectThenXRayCentreComposite,
    sim_run_engine: RunEngineSimulator,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
    test_config_files,
):
    mock_detect_grid_and_do_gridscan.return_value = iter(
        [Msg("detect_grid_and_do_gridscan")]
    )
    mock_pin_tip_centre_plan.return_value = iter([Msg("pin_tip_centre_plan")])

    msgs = sim_run_engine.simulate_plan(
        pin_centre_then_flyscan_plan(
            grid_detect_devices,
            test_pin_centre_then_xray_centre_params,
            test_config_files["oav_config_json"],
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "backlight"
        and msg.args == (BacklightPosition.IN,)
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "aperture_scatterguard"
        and msg.args == (ApertureValue.OUT_OF_BEAM,)
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "pin_tip_centre_plan"
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
def test_pin_centre_then_xray_centre_plan_goes_to_the_starting_chi_and_phi(
    mock_detect_grid_and_do_gridscan,
    mock_pin_tip_centre_plan,
    smargon: Smargon,
    sim_run_engine: RunEngineSimulator,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
    test_config_files,
    grid_detect_devices,
):
    mock_detect_grid_and_do_gridscan.return_value = iter(
        [Msg("detect_grid_and_do_gridscan")]
    )
    mock_pin_tip_centre_plan.return_value = iter([Msg("pin_tip_centre_plan")])

    mock_composite = grid_detect_devices
    mock_composite.smargon = smargon

    test_pin_centre_then_xray_centre_params.phi_start_deg = 30
    test_pin_centre_then_xray_centre_params.chi_start_deg = 50

    msgs = sim_run_engine.simulate_plan(
        pin_centre_then_flyscan_plan(
            grid_detect_devices,
            test_pin_centre_then_xray_centre_params,
            test_config_files["oav_config_json"],
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "smargon-phi"
        and msg.args == (test_pin_centre_then_xray_centre_params.phi_start_deg,)
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "smargon-chi"
        and msg.args == (test_pin_centre_then_xray_centre_params.chi_start_deg,)
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "pin_tip_centre_plan"
    )


def test_pin_tip_centre_then_xray_centre_fails_with_exception_when_no_beamstop(
    sim_run_engine: RunEngineSimulator,
    grid_detect_devices: GridDetectThenXRayCentreComposite,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
):
    sim_run_engine.add_read_handler_for(
        grid_detect_devices.beamstop.selected_pos, BeamstopPositions.UNKNOWN
    )
    with pytest.raises(BeamstopException):
        sim_run_engine.simulate_plan(
            pin_tip_centre_then_xray_centre(
                grid_detect_devices, test_pin_centre_then_xray_centre_params
            )
        )



import os
from contextlib import contextmanager
from pathlib import Path
from typing import Literal
from unittest.mock import patch

import h5py
import numpy as np
import pytest
from dodal.devices.detector.det_dim_constants import (
    PIXELS_X_EIGER2_X_4M,
    PIXELS_Y_EIGER2_X_4M,
)
from dodal.devices.fast_grid_scan import (
    GridAxis,
    ZebraGridScanParams,
)

from mx_bluesky.common.external_interaction.nexus.nexus_utils import (
    AxisDirection,
    create_beam_and_attenuator_parameters,
)
from mx_bluesky.common.external_interaction.nexus.write_nexus import NexusWriter
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

"""It's hard to effectively unit test the nexus writing so these are really system tests
that confirms that we're passing the right sorts of data to nexgen to get a sensible output.
Note that the testing process does now write temporary files to disk."""

TEST_FLUX = 1e10


def assert_end_data_correct(nexus_writer: NexusWriter):
    for filename in [nexus_writer.nexus_file, nexus_writer.master_file]:
        with h5py.File(filename, "r") as written_nexus_file:
            entry = written_nexus_file["entry"]
            assert isinstance(entry, h5py.Group)
            assert "end_time_estimated" in entry


def create_nexus_writer(parameters: HyperionSpecifiedThreeDGridScan, writer_num):
    d_size = parameters.detector_params.detector_size_constants.det_size_pixels
    n_img = (
        parameters.scan_indices[1]
        if writer_num == 1
        else parameters.num_images - parameters.scan_indices[1]
    )
    points = (
        parameters.scan_points_first_grid
        if writer_num == 1
        else parameters.scan_points_second_grid
    )
    data_shape = (n_img, d_size.width, d_size.height)
    run_number = parameters.detector_params.run_number + writer_num - 1
    vds_start = 0 if writer_num == 1 else parameters.scan_indices[1]
    omega_start = (
        parameters.grid1_omega_deg if writer_num == 1 else parameters.grid2_omega_deg
    )
    nexus_writer = NexusWriter(
        parameters,
        data_shape,
        scan_points=points,
        run_number=run_number,
        vds_start_index=vds_start,
        omega_start_deg=omega_start,
    )
    nexus_writer.beam, nexus_writer.attenuator = create_beam_and_attenuator_parameters(
        20, TEST_FLUX, 0.5
    )
    return nexus_writer


@contextmanager
def create_nexus_writers(parameters: HyperionSpecifiedThreeDGridScan):
    writers = [create_nexus_writer(parameters, i) for i in [1, 2]]
    writers[1].start_index = parameters.scan_indices[1]
    try:
        yield writers
    finally:
        for writer in writers:
            for file in [writer.nexus_file, writer.master_file]:
                if os.path.isfile(file):
                    os.remove(file)


@pytest.fixture
def dummy_nexus_writers(test_fgs_params: HyperionSpecifiedThreeDGridScan):
    with create_nexus_writers(test_fgs_params) as (
        nexus_writer_1,
        nexus_writer_2,
    ):
        yield nexus_writer_1, nexus_writer_2


@pytest.fixture
def dummy_nexus_writers_with_more_images(
    test_fgs_params: HyperionSpecifiedThreeDGridScan,
):
    x, y, z = 45, 35, 25
    test_fgs_params.x_steps = x
    test_fgs_params.y_steps = y
    test_fgs_params.z_steps = z
    with create_nexus_writers(test_fgs_params) as (
        nexus_writer_1,
        nexus_writer_2,
    ):
        yield nexus_writer_1, nexus_writer_2


@pytest.fixture
def single_dummy_file(test_fgs_params: HyperionSpecifiedThreeDGridScan):
    test_fgs_params.use_roi_mode = True
    d_size = test_fgs_params.detector_params.detector_size_constants.det_size_pixels
    data_shape = (test_fgs_params.scan_indices[1], d_size.width, d_size.height)
    nexus_writer = NexusWriter(
        test_fgs_params,
        data_shape,
        scan_points=test_fgs_params.scan_points_first_grid,
        run_number=1,
    )
    yield nexus_writer
    for file in [nexus_writer.nexus_file, nexus_writer.master_file]:
        if os.path.isfile(file):
            os.remove(file)


@pytest.mark.parametrize(
    "test_fgs_params, expected_num_of_files",
    [(2550, 3), (4000, 4), (8975, 9)],
    indirect=["test_fgs_params"],
)
def test_given_number_of_images_above_1000_then_expected_datafiles_used(
    test_fgs_params: HyperionSpecifiedThreeDGridScan,
    expected_num_of_files: Literal[3, 4, 9],
    single_dummy_file: NexusWriter,
):
    datafiles = single_dummy_file.get_image_datafiles()
    assert len(datafiles) == expected_num_of_files
    paths = [str(filename) for filename in datafiles]
    test_data_folder = (
        os.path.dirname(os.path.realpath(os.path.join(__file__, ".."))) + "/test_data"
    )
    expected_paths = [
        f"{test_data_folder}/dummy_1_00000{i + 1}.h5"
        for i in range(expected_num_of_files)
    ]
    assert expected_paths[0] == paths[0]


def test_given_dummy_data_then_datafile_written_correctly(
    test_fgs_params: HyperionSpecifiedThreeDGridScan,
    dummy_nexus_writers: tuple[NexusWriter, NexusWriter],
):
    nexus_writer_1, nexus_writer_2 = dummy_nexus_writers
    grid_scan_params: ZebraGridScanParams = test_fgs_params.FGS_params
    nexus_writer_1.create_nexus_file(np.uint16)

    for filename in [nexus_writer_1.nexus_file, nexus_writer_1.master_file]:
        with h5py.File(filename, "r") as written_nexus_file:
            assert isinstance(
                data_path := written_nexus_file["/entry/data"], h5py.Group
            )
            assert_x_data_stride_correct(
                data_path, grid_scan_params, grid_scan_params.y_steps
            )
            assert isinstance(sam_y := data_path["sam_y"], h5py.Dataset)
            assert_varying_axis_stride_correct(
                sam_y[:], grid_scan_params, grid_scan_params.y_axis
            )
            assert_axis_data_fixed(
                written_nexus_file, "z", grid_scan_params.z1_start_mm
            )
            assert isinstance(
                flux := written_nexus_file["/entry/instrument/beam/total_flux"],
                h5py.Dataset,
            )
            assert flux[()] == TEST_FLUX
            assert_contains_external_link(data_path, "data_000001", "dummy_1_000001.h5")
            assert isinstance(omega := data_path["omega"], h5py.Dataset)
            assert np.all(omega[:] == 0.0)

            assert np.all(
                written_nexus_file["/entry/data/omega"].attrs.get("vector")
                == [
                    -1.0,
                    0.0,
                    0.0,
                ]
            )
            assert np.all(
                written_nexus_file["/entry/data/sam_x"].attrs.get("vector")
                == [
                    1.0,
                    0.0,
                    0.0,
                ]
            )
            assert np.all(
                written_nexus_file["/entry/data/sam_y"].attrs.get("vector")
                == [
                    0.0,
                    1.0,
                    0.0,
                ]
            )

    assert_data_edge_at(nexus_writer_1.nexus_file, 629)
    assert_end_data_correct(nexus_writer_1)

    nexus_writer_2.create_nexus_file(np.uint16)

    for filename in [nexus_writer_2.nexus_file, nexus_writer_2.master_file]:
        with h5py.File(filename, "r") as written_nexus_file:
            assert isinstance(
                data_path := written_nexus_file["/entry/data"], h5py.Group
            )
            assert_x_data_stride_correct(
                data_path, grid_scan_params, grid_scan_params.z_steps
            )
            assert isinstance(sam_z := data_path["sam_z"], h5py.Dataset)
            assert_varying_axis_stride_correct(
                sam_z[:], grid_scan_params, grid_scan_params.z_axis
            )
            assert_axis_data_fixed(
                written_nexus_file, "y", grid_scan_params.y2_start_mm
            )
            assert isinstance(
                flux := written_nexus_file["/entry/instrument/beam/total_flux"],
                h5py.Dataset,
            )
            assert flux[()] == TEST_FLUX
            assert_contains_external_link(data_path, "data_000001", "dummy_1_000001.h5")
            assert_contains_external_link(data_path, "data_000002", "dummy_1_000002.h5")
            assert isinstance(omega := data_path["omega"], h5py.Dataset)
            assert np.all(omega[:] == 90.0)
            assert np.all(
                written_nexus_file["/entry/data/sam_z"].attrs.get("vector")
                == [
                    0.0,
                    0.0,
                    1.0,
                ]
            )

    assert_data_edge_at(nexus_writer_2.nexus_file, 419)


@pytest.mark.parametrize(
    "axis_direction, expected_vector",
    [[AxisDirection.NEGATIVE, [-1, 0, 0]], [AxisDirection.POSITIVE, [1, 0, 0]]],
)
def test_nexus_file_entry_data_omega_written_correctly_independent_of_omega_direction(
    test_rotation_params,
    axis_direction: AxisDirection,
    expected_vector: list[float],
    tmp_path: Path,
):
    test_rotation_params.storage_directory = str(tmp_path)
    det_size = (
        test_rotation_params.detector_params.detector_size_constants.det_size_pixels
    )
    shape = (test_rotation_params.num_images, det_size.width, det_size.height)
    nexus_writer = NexusWriter(
        test_rotation_params,
        shape,
        test_rotation_params.scan_points,
        omega_start_deg=test_rotation_params.omega_start_deg,
        chi_start_deg=test_rotation_params.chi_start_deg or 0,
        phi_start_deg=test_rotation_params.phi_start_deg or 0,
        vds_start_index=0,
        meta_data_run_number=1,
        axis_direction=axis_direction,
    )
    nexus_writer.beam, nexus_writer.attenuator = create_beam_and_attenuator_parameters(
        20, TEST_FLUX, 0.5
    )

    nexus_writer.create_nexus_file(np.uint16)
    with h5py.File(nexus_writer.nexus_file, "r") as nexus_file:
        assert all(
            nexus_file["/entry/data/omega"].attrs.get("vector") == expected_vector
        )
        data = nexus_file["/entry/data/omega"][:]  # type: ignore
        assert all(data == test_rotation_params.scan_points["omega"])


def assert_x_data_stride_correct(
    data_path, grid_scan_params: ZebraGridScanParams, varying_axis_steps
):
    sam_x_data = data_path["sam_x"][:]
    assert len(sam_x_data) == (grid_scan_params.x_steps) * (varying_axis_steps)
    assert sam_x_data[1] - sam_x_data[0] == pytest.approx(
        grid_scan_params.x_step_size_mm * 1000
    )


def assert_varying_axis_stride_correct(
    axis_data, grid_scan_params: ZebraGridScanParams, varying_axis: GridAxis
):
    assert len(axis_data) == (grid_scan_params.x_steps) * (varying_axis.full_steps)
    assert axis_data[grid_scan_params.x_steps + 1] - axis_data[0] == pytest.approx(
        varying_axis.step_size_mm * 1000
    )


def assert_axis_data_fixed(written_nexus_file, axis, expected_value):
    if f"sam_{axis}" in written_nexus_file["/entry/data"]:
        fixed_data = written_nexus_file["/entry/data"][f"sam_{axis}"][:]
        assert sum(fixed_data) == len(fixed_data) * fixed_data[0]


def assert_data_edge_at(nexus_file, expected_edge_index):
    """Asserts that the datafile's last datapoint is at the specified index"""
    with h5py.File(nexus_file) as f:
        assert isinstance(data := f["entry/data/data"], h5py.Dataset)
        assert data[expected_edge_index, 0, 0] == 0
        with pytest.raises(IndexError):
            assert data[expected_edge_index + 1, 0, 0] == 0


def assert_contains_external_link(data_path, entry_name, file_name):
    assert entry_name in data_path
    assert data_path[entry_name].file.filename.endswith(file_name)


def test_nexus_writer_files_are_formatted_as_expected(
    test_fgs_params: HyperionSpecifiedThreeDGridScan, single_dummy_file: NexusWriter
):
    for file in [single_dummy_file.nexus_file, single_dummy_file.master_file]:
        file_name = os.path.basename(file.name)
        expected_file_name_prefix = test_fgs_params.file_name + "_1"
        assert file_name.startswith(expected_file_name_prefix)


def test_nexus_writer_writes_width_and_height_correctly(single_dummy_file: NexusWriter):
    assert len(single_dummy_file.detector.detector_params.image_size) >= 2
    assert (
        single_dummy_file.detector.detector_params.image_size[0] == PIXELS_Y_EIGER2_X_4M
    )
    assert (
        single_dummy_file.detector.detector_params.image_size[1] == PIXELS_X_EIGER2_X_4M
    )


@patch.dict(os.environ, {"BEAMLINE": "I03"})
def test_nexus_writer_writes_beamline_name_correctly(
    test_fgs_params: HyperionSpecifiedThreeDGridScan,
):
    d_size = test_fgs_params.detector_params.detector_size_constants.det_size_pixels
    data_shape = (test_fgs_params.num_images, d_size.width, d_size.height)
    nexus_writer = NexusWriter(test_fgs_params, data_shape, test_fgs_params.scan_points)
    assert nexus_writer.source.beamline == "I03"


def check_validity_through_zocalo(nexus_writers: tuple[NexusWriter, NexusWriter]):
    import dlstbx.swmr.h5check

    nexus_writer_1, nexus_writer_2 = nexus_writers

    nexus_writer_1.create_nexus_file(np.uint16)
    nexus_writer_2.create_nexus_file(np.uint16)

    for filename in [nexus_writer_1.nexus_file, nexus_writer_1.master_file]:
        with h5py.File(filename, "r") as written_nexus_file:
            dlstbx.swmr.h5check.get_real_frames(
                written_nexus_file, written_nexus_file["entry/data/data"]
            )

    for filename in [nexus_writer_2.nexus_file, nexus_writer_2.master_file]:
        with h5py.File(filename, "r") as written_nexus_file:
            dlstbx.swmr.h5check.get_real_frames(
                written_nexus_file, written_nexus_file["entry/data/data"]
            )


@pytest.mark.dlstbx
def test_nexus_file_validity_for_zocalo_with_two_linked_datasets(
    dummy_nexus_writers: tuple[NexusWriter, NexusWriter],
):
    check_validity_through_zocalo(dummy_nexus_writers)


@pytest.mark.dlstbx
def test_nexus_file_validity_for_zocalo_with_three_linked_datasets(
    dummy_nexus_writers_with_more_images: tuple[NexusWriter, NexusWriter],
):
    check_validity_through_zocalo(dummy_nexus_writers_with_more_images)


@pytest.mark.skip("Requires #87 of nexgen")
def test_given_some_datafiles_outside_of_VDS_range_THEN_they_are_not_in_nexus_file(
    dummy_nexus_writers_with_more_images: tuple[NexusWriter, NexusWriter],
):
    nexus_writer_1, nexus_writer_2 = dummy_nexus_writers_with_more_images

    nexus_writer_1.create_nexus_file(np.uint16)
    nexus_writer_2.create_nexus_file(np.uint16)

    for filename in [nexus_writer_1.nexus_file, nexus_writer_1.master_file]:
        with h5py.File(filename, "r") as written_nexus_file:
            assert isinstance(data := written_nexus_file["entry/data"], h5py.Dataset)
            assert "data_000001" in data
            assert "data_000002" in data
            assert "data_000003" not in data

    for filename in [nexus_writer_2.nexus_file, nexus_writer_2.master_file]:
        with h5py.File(filename, "r") as written_nexus_file:
            assert isinstance(data := written_nexus_file["entry/data"], h5py.Dataset)
            assert "data_000001" not in data
            assert "data_000002" in data
            assert "data_000003" in data


def test_given_data_files_not_yet_written_when_nexus_files_created_then_nexus_files_still_written(
    test_fgs_params: HyperionSpecifiedThreeDGridScan,
):
    test_fgs_params.file_name = "non_existant_file"
    with create_nexus_writers(test_fgs_params) as (
        nexus_writer_1,
        nexus_writer_2,
    ):
        nexus_writer_1.create_nexus_file(np.uint16)
        nexus_writer_2.create_nexus_file(np.uint16)

        for filename in [
            nexus_writer_1.nexus_file,
            nexus_writer_1.master_file,
            nexus_writer_1.nexus_file,
            nexus_writer_1.master_file,
        ]:
            assert os.path.exists(filename)



from __future__ import annotations

import dataclasses
from collections.abc import Callable, Sequence
from functools import partial

import numpy as np
from bluesky.callbacks import CallbackBase
from event_model import RunStart

from mx_bluesky.common.parameters.components import (
    MultiXtalSelection,
    TopNByMaxCountSelection,
)


class XRayCentreEventHandler(CallbackBase):
    def __init__(self):
        super().__init__()
        self.xray_centre_results: Sequence[XRayCentreResult] | None = None

    def start(self, doc: RunStart) -> RunStart | None:
        if "xray_centre_results" in doc:
            self.xray_centre_results = [
                XRayCentreResult(**result_dict)
                for result_dict in doc["xray_centre_results"]  # type: ignore
            ]
        return doc


@dataclasses.dataclass
class XRayCentreResult:
    """
    Represents information about a hit from an X-ray centring.

    Attributes:
        centre_of_mass_mm: coordinates in mm of the centre of mass
        bounding_box_mm: coordinates in mm of opposite corners of the bounding box
            containing the crystal
        max_count: The maximum spot count encountered in any one grid box in the crystal
        total_count: The total count across all boxes in the crystal.
    """

    centre_of_mass_mm: np.ndarray
    bounding_box_mm: tuple[np.ndarray, np.ndarray]
    max_count: int
    total_count: int

    def __eq__(self, o):
        return (
            isinstance(o, XRayCentreResult)
            and o.max_count == self.max_count
            and o.total_count == self.total_count
            and all(o.centre_of_mass_mm == self.centre_of_mass_mm)
            and all(o.bounding_box_mm[0] == self.bounding_box_mm[0])
            and all(o.bounding_box_mm[1] == self.bounding_box_mm[1])
        )


def top_n_by_max_count(
    unfiltered: Sequence[XRayCentreResult], n: int
) -> Sequence[XRayCentreResult]:
    sorted_hits = sorted(unfiltered, key=lambda result: result.max_count, reverse=True)
    return sorted_hits[:n]


def resolve_selection_fn(
    params: MultiXtalSelection,
) -> Callable[[Sequence[XRayCentreResult]], Sequence[XRayCentreResult]]:
    if isinstance(params, TopNByMaxCountSelection):
        return partial(top_n_by_max_count, n=params.n)
    raise ValueError(f"Invalid selection function {params.name}")



import dataclasses
from typing import Any, ClassVar, Protocol, TypeVar, get_type_hints

from blueapi.core import BlueskyContext
from blueapi.core.bluesky_types import Device

from mx_bluesky.common.utils.log import LOGGER

T = TypeVar("T", bound=Device)


class _IsDataclass(Protocol):
    """Protocol followed by any dataclass"""

    __dataclass_fields__: ClassVar[dict]


DT = TypeVar("DT", bound=_IsDataclass)


def find_device_in_context(
    context: BlueskyContext,
    name: str,
    # Typing in here is wrong (see https://github.com/microsoft/pyright/issues/7228#issuecomment-1934500232)
    # but this whole thing will go away when we do https://github.com/DiamondLightSource/hyperion/issues/868
    expected_type: type[T] = Device,  # type: ignore
) -> T:
    LOGGER.debug(f"Looking for device {name} of type {expected_type} in context")

    device = context.find_device(name)
    if device is None:
        raise ValueError(
            f"Cannot find device named '{name}' in bluesky context {context.devices}."
        )

    if not isinstance(device, expected_type):
        raise ValueError(
            f"Found device named '{name}' and expected it to be a '{expected_type}' but it was a '{device.__class__.__name__}'"
        )

    LOGGER.debug(f"Found matching device {device}")
    return device


def device_composite_from_context(context: BlueskyContext, dc: type[DT]) -> DT:
    """
    Initializes all of the devices referenced in a given dataclass from a provided
    context, checking that the types of devices returned by the context are compatible
    with the type annotations of the dataclass.

    Note that if the context was not created with `wait_for_connection=True` devices may
    still be unconnected.
    """
    LOGGER.debug(
        f"Attempting to initialize devices referenced in dataclass {dc} from blueapi context"
    )

    devices: dict[str, Any] = {}
    dc_type_hints: dict[str, Any] = get_type_hints(dc)

    for field in dataclasses.fields(dc):
        device = find_device_in_context(
            context, field.name, expected_type=dc_type_hints.get(field.name, Device)
        )

        devices[field.name] = device

    return dc(**devices)



from __future__ import annotations

from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.detector.det_dim_constants import EIGER2_X_9M_SIZE, EIGER2_X_16M_SIZE
from dodal.devices.detector.detector import DetectorParams
from dodal.devices.fast_grid_scan import (
    ZebraGridScanParams,
)
from dodal.utils import get_beamline_name
from pydantic import Field, PrivateAttr
from scanspec.core import Path as ScanPath
from scanspec.specs import Line, Static

from mx_bluesky.common.parameters.components import (
    DiffractionExperimentWithSample,
    IspybExperimentType,
    OptionalGonioAngleStarts,
    SplitScan,
    WithOptionalEnergyChange,
    WithScan,
    XyzStarts,
)
from mx_bluesky.common.parameters.constants import (
    DetectorParamConstants,
    GridscanParamConstants,
    HardwareConstants,
)

DETECTOR_SIZE_PER_BEAMLINE = {"i02-1": EIGER2_X_9M_SIZE, "dev": EIGER2_X_16M_SIZE}


class GridCommon(
    DiffractionExperimentWithSample,
    OptionalGonioAngleStarts,
):
    """
    Parameters used in every MX diffraction experiment using grids. This model should
    be used by plans which have no knowledge of the grid specifications - i.e before
    automatic grid detection has completed
    """

    box_size_um: float = Field(default=GridscanParamConstants.BOX_WIDTH_UM)
    grid_width_um: float = Field(default=GridscanParamConstants.WIDTH_UM)
    exposure_time_s: float = Field(default=GridscanParamConstants.EXPOSURE_TIME_S)

    ispyb_experiment_type: IspybExperimentType = Field(
        default=IspybExperimentType.GRIDSCAN_3D
    )
    selected_aperture: ApertureValue | None = Field(default=ApertureValue.SMALL)

    tip_offset_um: float = Field(default=HardwareConstants.TIP_OFFSET_UM)


class SpecifiedGrid(XyzStarts, WithScan):
    """A specified grid is one which has defined values for the start position,
    grid and box sizes, etc., as opposed to parameters for a plan which will create
    those parameters at some point (e.g. through optical pin detection)."""


class SpecifiedThreeDGridScan(
    GridCommon,
    SpecifiedGrid,
    SplitScan,
    WithOptionalEnergyChange,
):
    """Parameters representing a so-called 3D grid scan, which consists of doing a
    gridscan in X and Y, followed by one in X and Z."""

    grid1_omega_deg: float = Field(default=GridscanParamConstants.OMEGA_1)  # type: ignore
    grid2_omega_deg: float = Field(default=GridscanParamConstants.OMEGA_2)
    x_step_size_um: float = Field(default=GridscanParamConstants.BOX_WIDTH_UM)
    y_step_size_um: float = Field(default=GridscanParamConstants.BOX_WIDTH_UM)
    z_step_size_um: float = Field(default=GridscanParamConstants.BOX_WIDTH_UM)
    y2_start_um: float
    z2_start_um: float
    x_steps: int = Field(gt=0)
    y_steps: int = Field(gt=0)
    z_steps: int = Field(gt=0)
    _set_stub_offsets: bool = PrivateAttr(default_factory=lambda: False)

    @property
    def FGS_params(self) -> ZebraGridScanParams:
        return ZebraGridScanParams(
            x_steps=self.x_steps,
            y_steps=self.y_steps,
            z_steps=self.z_steps,
            x_step_size_mm=self.x_step_size_um / 1000,
            y_step_size_mm=self.y_step_size_um / 1000,
            z_step_size_mm=self.z_step_size_um / 1000,
            x_start_mm=self.x_start_um / 1000,
            y1_start_mm=self.y_start_um / 1000,
            z1_start_mm=self.z_start_um / 1000,
            y2_start_mm=self.y2_start_um / 1000,
            z2_start_mm=self.z2_start_um / 1000,
            set_stub_offsets=self._set_stub_offsets,
            dwell_time_ms=self.exposure_time_s * 1000,
            transmission_fraction=self.transmission_frac,
        )

    def do_set_stub_offsets(self, value: bool):
        self._set_stub_offsets = value

    @property
    def grid_1_spec(self):
        x_end = self.x_start_um + self.x_step_size_um * (self.x_steps - 1)
        y1_end = self.y_start_um + self.y_step_size_um * (self.y_steps - 1)
        grid_1_x = Line("sam_x", self.x_start_um, x_end, self.x_steps)
        grid_1_y = Line("sam_y", self.y_start_um, y1_end, self.y_steps)
        grid_1_z = Static("sam_z", self.z_start_um)
        return grid_1_y.zip(grid_1_z) * ~grid_1_x

    @property
    def grid_2_spec(self):
        x_end = self.x_start_um + self.x_step_size_um * (self.x_steps - 1)
        z2_end = self.z2_start_um + self.z_step_size_um * (self.z_steps - 1)
        grid_2_x = Line("sam_x", self.x_start_um, x_end, self.x_steps)
        grid_2_z = Line("sam_z", self.z2_start_um, z2_end, self.z_steps)
        grid_2_y = Static("sam_y", self.y2_start_um)
        return grid_2_z.zip(grid_2_y) * ~grid_2_x

    @property
    def scan_indices(self):
        """The first index of each gridscan, useful for writing nexus files/VDS"""
        return [
            0,
            len(ScanPath(self.grid_1_spec.calculate()).consume().midpoints["sam_x"]),
        ]

    @property
    def scan_spec(self):
        """A fully specified ScanSpec object representing both grids, with x, y, z and
        omega positions."""
        return self.grid_1_spec.concat(self.grid_2_spec)

    @property
    def scan_points(self):
        """A list of all the points in the scan_spec."""
        return ScanPath(self.scan_spec.calculate()).consume().midpoints

    @property
    def scan_points_first_grid(self):
        """A list of all the points in the first grid scan."""
        return ScanPath(self.grid_1_spec.calculate()).consume().midpoints

    @property
    def scan_points_second_grid(self):
        """A list of all the points in the second grid scan."""
        return ScanPath(self.grid_2_spec.calculate()).consume().midpoints

    @property
    def num_images(self) -> int:
        return len(self.scan_points["sam_x"])

    @property
    def detector_params(self):
        self.det_dist_to_beam_converter_path = (
            self.det_dist_to_beam_converter_path
            or DetectorParamConstants.BEAM_XY_LUT_PATH
        )
        optional_args = {}
        if self.run_number:
            optional_args["run_number"] = self.run_number
        assert self.detector_distance_mm is not None, (
            "Detector distance must be filled before generating DetectorParams"
        )
        return DetectorParams(
            detector_size_constants=DETECTOR_SIZE_PER_BEAMLINE[
                get_beamline_name("dev")
            ],
            expected_energy_ev=self.demand_energy_ev,
            exposure_time_s=self.exposure_time_s,
            directory=self.storage_directory,
            prefix=self.file_name,
            detector_distance=self.detector_distance_mm,
            omega_start=self.omega_start_deg or 0,
            omega_increment=0,
            num_images_per_trigger=1,
            num_triggers=self.num_images,
            use_roi_mode=self.use_roi_mode,
            det_dist_to_beam_converter_path=self.det_dist_to_beam_converter_path,
            trigger_mode=self.trigger_mode,
            **optional_args,
        )



# from collections.abc import Sequence
from typing import Literal

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from blueapi.core import MsgGenerator
from dodal.beamlines import i24

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import (
    ChipType,
    MappingType,
    PumpProbeSetting,
)
from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick import (
    _move_on_mouse_click_plan,
)
from mx_bluesky.beamlines.i24.serial.log import _read_visit_directory_from_file
from mx_bluesky.beamlines.i24.serial.parameters import (
    FixedTargetParameters,
    get_chip_format,
)
from mx_bluesky.beamlines.i24.serial.setup_beamline import pv
from mx_bluesky.beamlines.i24.serial.setup_beamline.ca import caput
from mx_bluesky.beamlines.i24.serial.setup_beamline.pv_abstract import Eiger, Pilatus
from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector import (
    _move_detector_stage,
    get_detector_type,
)


@bpp.run_decorator()
def gui_stage_move_on_click(position_px: tuple[int, int]) -> MsgGenerator:
    oav = i24.oav()
    pmac = i24.pmac()
    yield from _move_on_mouse_click_plan(oav, pmac, position_px)


@bpp.run_decorator()
def gui_gonio_move_on_click(position_px: tuple[int, int]) -> MsgGenerator:
    oav = i24.oav()
    gonio = i24.vgonio()

    x_pixels_per_micron = yield from bps.rd(oav.microns_per_pixel_x)
    y_pixels_per_micron = yield from bps.rd(oav.microns_per_pixel_y)

    x_um = position_px[0] * x_pixels_per_micron
    y_um = position_px[1] * y_pixels_per_micron

    # gonio is in mm?
    yield from bps.mv(gonio.x, x_um / 1000, gonio.yh, y_um / 1000)  # type: ignore


# See https://github.com/DiamondLightSource/mx-bluesky/issues/853
@bpp.run_decorator()
def gui_sleep(sec: int) -> MsgGenerator:
    for _ in range(sec):
        yield from bps.sleep(1)


@bpp.run_decorator()
def gui_move_detector(det: Literal["eiger", "pilatus"]) -> MsgGenerator:
    detector_stage = i24.detector_motion()
    det_y_target = Eiger.det_y_target if det == "eiger" else Pilatus.det_y_target
    yield from _move_detector_stage(detector_stage, det_y_target)
    # Make the output readable
    caput(pv.me14e_gp101, det)


@bpp.run_decorator()
def gui_set_parameters(
    sub_dir: str,
    chip_name: str,
    exp_time: float,
    det_dist: float,
    transmission: float,
    n_shots: int,
    chip_type: str,
    checker_pattern: bool,
    pump_probe: str,
    laser_dwell: float,
    laser_delay: float,
    pre_pump: float,
) -> MsgGenerator:
    # NOTE still a work in progress, adding to it as the ui grows
    detector_stage = i24.detector_motion()
    det_type = yield from get_detector_type(detector_stage)
    chip_params = get_chip_format(ChipType[chip_type])

    params = {
        "visit": _read_visit_directory_from_file().as_posix(),  # noqa
        "directory": sub_dir,
        "filename": chip_name,
        "exposure_time_s": exp_time,
        "detector_distance_mm": det_dist,
        "detector_name": str(det_type),
        "num_exposures": n_shots,
        "transmission": transmission,
        "chip": chip_params,
        "map_type": MappingType.NoMap,
        "chip_map": [],
        "pump_repeat": PumpProbeSetting[pump_probe],  # pump_repeat,
        "laser_dwell_s": laser_dwell,
        "laser_delay_s": laser_delay,
        "checker_pattern": checker_pattern,
        "pre_pump_exposure_s": pre_pump,
    }
    print(FixedTargetParameters(**params))
    # This will then run the run_fixed_target plan
    yield from bps.sleep(0.5)
